#! /usr/bin/env python3

import sys, gzip, statistics, logging
import pysam

from nanomonsv.logger import get_logger

logger = get_logger(__name__)

class Sbnd_cluster(object):

    def __init__(self, tchr, tstart, tend, tdir, treadid, tinfo):

        self.chr = tchr
        self.start = tstart
        self.end = tend
        self.dir = tdir
        self.readids = [treadid]
        self.info = [tinfo]


class Sbnd_clusterer(object):

    def __init__(self, output_file, control_bed = None, control_panel_bed = None,
        cluster_margin_size = None, read_num_thres = None, median_mapQ_thres = None,
        max_control_read_num = None, max_panel_read_num = None, max_panel_sample_num = None):

        self.hout = open(output_file, 'w')
        self.control_tb = None
        if control_bed is not None:
            self.control_tb = pysam.TabixFile(control_bed)

        self.control_panel_tb = None
        if control_panel_bed is not None:
            self.control_panel_tb = pysam.TabixFile(control_panel_bed)

        self.sbnd_cluster_list = []
        self.cluster_margin_size = cluster_margin_size
        self.read_num_thres = read_num_thres
        self.median_mapQ_thres = median_mapQ_thres
        self.max_control_read_num = max_control_read_num
        self.max_panel_read_num = max_panel_read_num
        self.max_panel_sample_num = max_panel_sample_num

    def __del__(self):
        self.hout.close()
        if self.control_tb is not None: self.control_tb.close()
        if self.control_panel_tb is not None: self.control_panel_tb.close()

 
    def check_mergeability(self, tchr, tstart, tend, tdir, treadid, tinfo):
   
        for cluster in self.sbnd_cluster_list:

            if treadid in cluster.readids: continue

            # pairs of chromosome and direction should be the same.
            # regions from start to end should overlap.   
            is_mergeable = False
            if tchr == cluster.chr and tdir == cluster.dir and tend >= cluster.start and tstart <= cluster.end:
                is_mergeable = True

            if not is_mergeable: continue

            cluster.start = min(tstart, cluster.start)
            cluster.end = max(tend, cluster.end)

            cluster.readids.append(treadid)
            cluster.info.append(tinfo)
                        
            return cluster
            
        return None


    def add_new_cluster(self, tchr, tstart, tend, tdir, treadid, tinfo):

        self.sbnd_cluster_list.append(
            Sbnd_cluster(tchr, tstart, tend, tdir, treadid, tinfo))


    def filter_single_breakend_cluster(self, cl):
        
        is_filter = False
        if len(cl.readids) < self.read_num_thres: is_filter = True

        median_mapQ = statistics.median([int(x.split(',')[5]) for x in cl.info])
        if median_mapQ < self.median_mapQ_thres: is_filter = True

        bp_size_vec = []
        for x in cl.info:
            xx = x.split(',')
            if xx[0] == xx[1]: bp_size_vec.append(int(xx[0]))
            if xx[1] == xx[2]: bp_size_vec.append(int(xx[3]) - int(xx[2]))

        if statistics.median(bp_size_vec) < 1000: is_filter = True

        if is_filter == True: return(True)

        control_flag = False
        if self.control_tb is not None:
    
            support_read_num = 0
            tabix_error_flag = False
            try:
                records = self.control_tb.fetch(cl.chr, max(0, cl.start - 10), cl.end + 10)
            except Exception as e:
                logger.debug(f'{e}')
                tabix_error_flag = True

            if not tabix_error_flag:
                for record_line in records:
                    rec = record_line.split('\t')

                    if cl.chr != rec[0] or cl.dir != rec[5]: continue
        
                    if cl.end >= int(rec[2]) and cl.start <= int(rec[2]):
                        support_read_num = support_read_num + 1

            if support_read_num > self.max_control_read_num:
                control_flag = True


        control_panel_flag = False
        if self.control_panel_tb is not None:
        
            tabix_error_flag = False
            try:
                records = self.control_panel_tb.fetch(cl.chr, max(0, cl.start - 10), cl.end + 10)
            except Exception as e:
                logger.debug(f'{e}')
                tabix_error_flag = True
        
            sample2readnum_panel = {}        
            if not tabix_error_flag:
                for record_line in records:
                    rec = record_line.split('\t')
                    
                    if cl.chr != rec[0] or cl.dir != rec[5]: continue
                    
                    if cl.end >= int(rec[2]) and cl.start <= int(rec[2]):
                        readnums = rec[6].split(',')
                        psamples = rec[7].split(',')
                        for psample, readnum in zip(psamples, readnums):
                            if psample not in sample2readnum_panel:
                                sample2readnum_panel[psample] = 0
                            sample2readnum_panel[psample] = sample2readnum_panel[psample] + int(readnum)

                readnums_panel = [sample2readnum_panel[x] for x in sample2readnum_panel]
                if len([x for x in readnums_panel if x > self.max_panel_read_num]) > self.max_panel_sample_num:
                    control_panel_flag = True
            

        if control_flag == True or control_panel_flag == True: is_filter = True 

        return(is_filter)


    def flush_sv_cluster_list(self, current_chr, current_pos):

        remove_cluster = []
        for i in range(len(self.sbnd_cluster_list)):
            cl = self.sbnd_cluster_list[i]
            if current_chr != cl.chr or current_pos > cl.end + self.cluster_margin_size:
                remove_cluster.append(cl)

                if self.filter_single_breakend_cluster(cl): continue
                # print_line_readids = ';'.join(cl.readids)
                print_line_readids = repr(cl.readids)
                print_line_info = ';'.join(cl.info)
        
                print(f'{cl.chr}\t{cl.start}\t{cl.end}\t{print_line_readids}\t0\t{cl.dir}\t{print_line_info}',
                    file = self.hout)
                        
        for cl in remove_cluster:
            self.sbnd_cluster_list.remove(cl)


def cluster_supporting_reads_sbnd(input_file, output_file, control_bed = None, control_panel_bed = None, 
    cluster_margin_size = 100, sbnd_cluster_margin_size = 20, 
    read_num_thres = 3, median_mapQ_thres = 20, max_control_read_num = 1,
    max_panel_read_num = 1, max_panel_sample_num = 1, debug = False):

    if debug: logger.setLevel(logging.DEBUG)


    sbnd_clusterer = Sbnd_clusterer(output_file, control_bed = control_bed, control_panel_bed = control_panel_bed,
        cluster_margin_size = cluster_margin_size, read_num_thres = read_num_thres, 
        median_mapQ_thres = median_mapQ_thres, max_control_read_num = max_control_read_num, 
        max_panel_read_num = max_panel_read_num, max_panel_sample_num = max_panel_sample_num)

    
    with gzip.open(input_file, 'rt') as hin: 
        for line in hin:

            tchr, _, tpos, treadid, _, tdir, tinfo = line.rstrip('\n').split('\t')
            tstart = max(int(tpos) - sbnd_cluster_margin_size, 0)
            tend = int(tpos) + sbnd_cluster_margin_size

            # skip single breakend from secondary alignment
            if tinfo.split(',')[9] == "True": continue

            # flush out existing cluster whose supporting reads have parsed alreadly 
            #   (considering the chromosome and coordinates).
            sbnd_clusterer.flush_sv_cluster_list(tchr, int(tpos))

            # check the new single breakend can be merged into any existing clusters
            cret = sbnd_clusterer.check_mergeability(tchr, tstart, tend, tdir, treadid, tinfo)

            # create new single breakend cluster when new key cannot be merged into any existing clusteres
            if cret is None:
                sbnd_clusterer.add_new_cluster(tchr, tstart, tend, tdir, treadid, tinfo)
 

        sbnd_clusterer.flush_sv_cluster_list("EOF", 0)


if __name__ == "__main__":

    import sys
    input_file = sys.argv[1]
    output_file = sys.argv[2]
    control = sys.argv[3]

    cluster_supporting_reads_single_breakend(input_file, output_file, control_bed = control)
             



#! /usr/bin/env python3

import sys, re, os, gzip, subprocess # pkg_resources
import pysam

from .swalign import *
from .my_seq import get_seq, reverse_complement

def make_fasta_file(input_file, output_file, seq_id_file):

    
    with open(input_file, 'r') as hin, open(output_file, 'w') as hout1, open(seq_id_file, 'w') as hout2:
        header = hin.readline()
        sid = 1
        for line in hin:
            F = line.rstrip('\n').split('\t')
            if len(F[6]) < 100: continue
        
            key = ','.join(F[:6] + [str(len(F[6]))])

            print(">seq" + str(sid) + ',' + str(len(F[6])), file = hout1)
            print(F[6], file = hout1)
            print("seq" + str(sid) + ',' + str(len(F[6])) + '\t' + key, file = hout2)
            sid = sid + 1
 

def make_exon_bed_from_gtf(gtf_file, output_file):

    if gtf_file.endswith(".gtf"):
        hin = open(gtf_file, 'r')
    elif gtf_file.endswith(".gtf.gz"):
        hin = gzip.open(gtf_file, 'rt')
    else:
        print("The input GTF file should ends with 'gtf' or 'gtf.gz'", file = sys.stderr)
        sys.exit(1)

    hout = open(output_file + ".unsorted.tmp", 'w')
    for line in hin:
        if line.startswith('#'): continue
        F = line.rstrip('\n').split('\t')
        if F[2] != "exon": continue
        infos = F[8].split('; ')

        transcript_id = None
        exon_number = None
        for elm in infos:
            if elm.startswith("transcript_id "):
                transcript_id = elm.replace("transcript_id ", '').strip('"')
            if elm.startswith("exon_number "):
                exon_number = elm.replace("exon_number ", '').strip('"')

        
        print(f'{F[0]}\t{int(F[3]) - 1}\t{F[4]}\t{transcript_id}\t{exon_number}\t{F[6]}', file = hout)

    hin.close()
    hout.close()

    with open(output_file + ".sorted.tmp", 'w') as hout:
        subprocess.run(["sort", "-k1,1", "-k2,2n", "-k3,3n", output_file + ".unsorted.tmp"], stdout = hout)
    
    with open(output_file, 'w') as hout:
        subprocess.run(["bgzip", "-f", "-c", output_file + ".sorted.tmp"], stdout = hout)

    subprocess.run(["tabix", "-p", "bed", output_file])

    os.remove(output_file + ".unsorted.tmp")
    os.remove(output_file + ".sorted.tmp")


def sam2bed_split(input_file, output_file):

    sam_file = pysam.AlignmentFile(input_file, 'r')
    hout = open(output_file, 'w')

    for read in sam_file.fetch():

        if read.is_unmapped: continue

        is_secondary = read.is_secondary
        is_supplementary = read.is_supplementary
        if read.is_secondary: continue

        query_name = read.query_name
        query_length = read.infer_read_length()
        query_strand = '-' if read.is_reverse else '+'

        reference_name = read.reference_name
        reference_start = read.reference_start + 1
        reference_end = read.reference_end

        cigartuples = read.cigartuples
        left_hard_clipping_size, right_hard_clipping_size = 0, 0
        if cigartuples[0][0] == 5: left_hard_clipping_size = cigartuples[0][1]
        if cigartuples[-1][0] == 5: right_hard_clipping_size = cigartuples[-1][1]
    
        if not is_supplementary:
            if query_strand == '+':
                query_start = read.query_alignment_start + 1 
                query_end = read.query_alignment_end
            else:
                query_start = query_length - read.query_alignment_end + 1
                query_end = query_length - read.query_alignment_start
        else:
            if query_strand == '+':
                query_start = left_hard_clipping_size + 1
                query_end = query_length - right_hard_clipping_size
            else:
                query_start = right_hard_clipping_size + 1
                query_end = query_length - left_hard_clipping_size

        query_pos_cur = query_start - 1 if query_strand == '+' else query_end
        query_pos_check = query_pos_cur
        query_exon_cur = query_pos_cur
        reference_pos_cur = reference_start - 1
        reference_exon_cur = reference_pos_cur
        reference_pos_check = reference_start - 1


        for cigar in cigartuples:
            if cigar[0] == 0:
                query_pos_cur = query_pos_cur + cigar[1] if query_strand == '+' else query_pos_cur - cigar[1]
                reference_pos_cur = reference_pos_cur + cigar[1] 
            elif cigar[0] == 1:
                query_pos_cur = query_pos_cur + cigar[1] if query_strand == '+' else query_pos_cur - cigar[1]
            elif cigar[0] == 2:
                reference_pos_cur = reference_pos_cur + cigar[1]
            elif cigar[0] == 3:
                exon_size = reference_pos_cur - reference_exon_cur 
                query_name2 = query_name + ',' + str(min(query_exon_cur, query_pos_cur) + 1) + ',' + \
                  str(max(query_exon_cur, query_pos_cur))
                print('\t'.join([reference_name, str(reference_exon_cur), str(reference_pos_cur), 
                                 query_name2, str(exon_size), query_strand]), file = hout)

                reference_pos_cur = reference_pos_cur + cigar[1]
                reference_exon_cur = reference_pos_cur
                query_exon_cur = query_pos_cur

        exon_size = reference_pos_cur - reference_exon_cur
        query_name2 = query_name + ',' + str(min(query_exon_cur, query_pos_cur) + 1) + ',' + \
        str(max(query_exon_cur, query_pos_cur))
        print('\t'.join([reference_name, str(reference_exon_cur), str(reference_pos_cur),
                         query_name2, str(exon_size), query_strand]), file = hout)

        if query_strand == '+' and query_end != query_pos_cur:
            print("query end inconsistent!!")
            sys.exit(1)
        if query_strand == '-' and query_start != query_pos_cur + 1:
            print("query end inconsistent!!")
            sys.exit(1)

    # read.qname = read.qname + ',' + str(query_start) + ',' + str(query_end) + ',' + query_strand
    hout.close()



def pp_proc_filt_exon(input_file, seq_list, output_file):

    key2aln_info = {}
    key2insert_len = {}
    key2total_match = {}
    key2exon_match_num = {}

    sid2skey = {}
    with open(seq_list, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            sid2skey[F[0]] = F[1]

    with open(input_file, 'r') as hin, open(output_file, 'w') as hout:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            insert_key_info = F[3].split(',')
            insert_key = ','.join(insert_key_info[:2])
            insert_len = int(insert_key_info[1])
            aln_start, aln_end = int(insert_key_info[2]), int(insert_key_info[3])

            gene = F[9]
            exon_num = F[10]
            key = insert_key + '\t' + gene
            key2insert_len[key] = insert_len
            if key not in key2total_match: key2total_match[key] = 0
            if key not in key2exon_match_num: key2exon_match_num[key] = 0
            if key not in key2aln_info: key2aln_info[key] = []

            key2total_match[key] = key2total_match[key] + min(int(F[12]), aln_end - aln_start + 1)

            match_ratio1 = float(F[12]) / (int(F[8]) - int(F[7]))
            match_ratio2 = float(F[12]) / (int(F[2]) - int(F[1]))
            if min(match_ratio1, match_ratio2) >= 0.95: key2exon_match_num[key] = key2exon_match_num[key] + 1

            if int(F[7]) - int(F[1]) > 10:
                if F[5] == '+':
                    aln_start = min(aln_end, aln_start + (int(F[7]) - int(F[1])))
                else:
                    aln_end = max(aln_start, aln_end - (int(F[7]) - int(F[1])))
            
            if int(F[2]) - int(F[8]) > 10:
                if F[5] == '+':
                    aln_end = max(aln_start, aln_end - (int(F[2]) - int(F[8])))
                else:
                    aln_start = min(aln_end, aln_start + (int(F[2]) - int(F[8])))

            key2aln_info[key].append((aln_start, aln_end, exon_num, round(match_ratio1, 4)))

        for key in key2aln_info:

            total_match_ratio = float(key2total_match[key]) / key2insert_len[key]

            aln_bar = []
            for ainfo in sorted(key2aln_info[key]):
                aln_bar.append(','.join([str(ainfo[0]), str(ainfo[1]), str(ainfo[2]), str(ainfo[3])]))

            if total_match_ratio < 0.5: continue
            if key2exon_match_num[key] < 2: continue

            print(key + '\t' + str(round(total_match_ratio, 4)) + '\t' + str(key2exon_match_num[key]) + '\t' + ';'.join(aln_bar), file = hout)
     


def proc_rmsk_info(tkey, rmsk_info):

    _, total_len = tkey.split(',')
    total_len = int(total_len)
    alu_len, alu_count = 0, 0
    L1_len, L1_count, L1_dir = 0, 0, []
    SVA_len, SVA_count = 0, 0
    repeat_len, repeat_count = 0, 0
    polyAT_len = 0

    repeat_info = []
    for i in range(len(rmsk_info)):

        F = rmsk_info[i]

        if F[8] == 'C': F[8] = '-'
        if F[8] == '+':
            repeat_key = ','.join([F[5], F[6], F[8], F[9], F[10], F[11], F[12]])
        else:
            repeat_key = ','.join([F[5], F[6], F[8], F[9], F[10], F[12], F[13]])
        repeat_info.append(repeat_key)

        if F[10] == "LINE/L1":
            L1_len = L1_len + int(F[6]) - int(F[5]) + 1
            L1_count = L1_count + 1
            L1_dir.append(F[8])
        if F[10] == "SINE/Alu":
            alu_len = alu_len + int(F[6]) - int(F[5]) + 1
            alu_count = alu_count + 1
        if F[10] == "Retroposon/SVA":
            SVA_len = SVA_len + int(F[6]) - int(F[5]) + 1
            SVA_count = SVA_count + 1
        if F[9] == "(T)n" and i == 0:
            polyAT_len = polyAT_len + int(F[6])   
        if F[9] == "(A)n" and i == len(rmsk_info) - 1:
            polyAT_len = polyAT_len + total_len - int(F[5]) + 1


    repeat_class = "None"
    L1_ratio, Alu_ratio, SVA_ratio = 0.0, 0.0, 0.0
    if float(total_len) - float(polyAT_len) > 0:
        L1_ratio = min(1.0, float(L1_len) / (float(total_len) - float(polyAT_len)))
        Alu_ratio = min(1.0, float(alu_len) / (float(total_len) - float(polyAT_len)))
        SVA_ratio = min(1.0, float(SVA_len) / (float(total_len) - float(polyAT_len)))


    if L1_ratio >= 0.8:
        if L1_count == 1:
            repeat_class = "Simple_LINE1"
        elif L1_count == 2 and L1_dir[0] != L1_dir[1]:
            repeat_class = "Inverted_LINE1"
        else:
            repeat_class = "Other_LINE1"

    if Alu_ratio >= 0.8:
        repeat_class = "Alu"

    if SVA_ratio >= 0.8:
        repeat_class = "SVA"

    return([repeat_class, str(round(L1_ratio, 4)), str(round(Alu_ratio, 4)), str(round(SVA_ratio, 4)), repeat_info])


def summarize_rmsk(input_file, output_file):

    temp_key = ''
    temp_rmsk_info = []

    with open(input_file, 'r') as hin, open(output_file, 'w') as hout:
        for line in hin:
            F = line.rstrip('\n').split()
            if len(F) <= 1 or F[0] in ["SW", "score"]: continue
            if F[0] == "There": continue

            if temp_key != F[4]:
                if temp_key != '':

                    repeat_class, L1_ratio, Alu_ratio, SVA_ratio, repeat_info = proc_rmsk_info(temp_key, temp_rmsk_info)
                    print(temp_key + '\t' + repeat_class + '\t' + L1_ratio + '\t' + Alu_ratio + '\t' + SVA_ratio + '\t' + ';'.join(repeat_info), file = hout)

                temp_key = F[4]
                temp_rmsk_info = []
            

            temp_rmsk_info.append(F)

        if temp_key != '':
            repeat_class, L1_ratio, Alu_ratio, SVA_ratio, repeat_info = proc_rmsk_info(temp_key, temp_rmsk_info)
            print(temp_key + '\t' + repeat_class + '\t' + L1_ratio + '\t' + Alu_ratio + '\t' + SVA_ratio + '\t' + ';'.join(repeat_info), file = hout)



def check_tsd_polyAT(input_file, seq_list, reference, output_file):

    match = 1
    mismatch = -8
    scoring = NucleotideScoringMatrix(match, mismatch)
    sw = LocalAlignment(scoring, gap_penalty=-8, gap_extension_penalty=-8)

    sid2skey = {}
    with open(seq_list, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            sid2skey[F[0]] = F[1]

    with open(input_file, 'r') as hin, open(output_file, 'w') as hout:
        for line in hin:
            sid = line.rstrip('\n').lstrip('>')
            seq = hin.readline().rstrip('\n')
            skey = sid2skey[sid]

            tchr, tpos1, _, _, tpos2, _, tlen = skey.split(',')

            match1 = re.search(r'^([ACGT]{1,20}?)TTTTTTTTTT', seq)
            match2 = re.search(r'AAAAAAAAAA([ACGT]{1,20}?)$', seq)
            tsd1 = None
            tsd2 = None
            is_polyA = False
            is_polyT = False

            if seq.startswith('TTTTTTTTTT') or match1 is not None: is_polyT = True
            if seq.endswith('AAAAAAAAAA') or match2 is not None: is_polyA = True


            tsd_cand1 = seq[:20]
            local_seq1 = get_seq(reference, tchr, int(tpos2), int(tpos2) + len(tsd_cand1))
            alignment1 = sw.align(tsd_cand1, local_seq1)
            if alignment1.q_pos <= 2 and alignment1.r_pos <= 2 and alignment1.matches >= 5 and alignment1.identity >= 0.8: 
                tsd1 = alignment1.query[(alignment1.q_pos):(alignment1.q_pos + alignment1.matches)]  


            tsd_cand2 = reverse_complement(seq[-20:])
            local_seq2 = reverse_complement(get_seq(reference, tchr, int(tpos1) - len(tsd_cand2), int(tpos1)))
            alignment2 = sw.align(tsd_cand2, local_seq2)
            if alignment2.q_pos <= 2 and alignment2.r_pos <= 2 and alignment2.matches >= 5 and alignment2.identity >= 0.8: 
                tsd2 = reverse_complement(alignment2.query[(alignment2.q_pos):(alignment2.q_pos + alignment2.matches)])

            if tsd1 is not None and tsd2 is None: 
                tsd = tsd1
            elif tsd1 is None and tsd2 is not None:
                tsd = tsd2
            elif tsd1 is not None and tsd2 is not None:
                tsd = tsd1 if len(tsd1) >= len(tsd2) else tsd2
            else:
                tsd = None

            polyAT = None
            if is_polyT: polyAT = "polyT"
            if is_polyA: polyAT = "polyA"

            # print(sid + '\t' + str(polyAT) + '\t' + str(tsd) + '\t' + tsd_cand1 + '\t' + local_seq1 + '\t' + tsd_cand2 + '\t' + local_seq2, file = hout) 
            print(f'{sid}\t{polyAT}\t{tsd}\t{tsd_cand1}\t{local_seq1}\t{alignment1.q_pos}\t{alignment1.r_pos}\t{alignment1.matches}\t{alignment1.identity}\t{tsd_cand2}\t{local_seq2}\t{alignment2.q_pos}\t{alignment2.r_pos}\t{alignment2.matches}\t{alignment2.identity}', file = hout)



def summarize_bwa_alignment(input_sam, seq_list, output_file):

    samfile = pysam.AlignmentFile(input_sam, 'r')

    inserted_positions = []
    with open(seq_list, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tchr, tstart, _, _, tend, _, _ = F[1].split(',')
            inserted_positions.append((tchr, int(tstart), int(tend)))


    samfile = pysam.AlignmentFile(input_sam, 'r')
    hout = open(output_file, 'w') 
    for read in samfile.fetch():

        if read.is_secondary or read.is_supplementary: continue
        mapping_quality = read.mapping_quality
        if mapping_quality < 30: continue
 
        query_name = read.query_name
        key = query_name

        query_strand = '-' if read.is_reverse else '+'
        reference_name = read.reference_name
        reference_start = read.reference_start + 1
        reference_end = read.reference_end
        mapping_quality = read.mapping_quality
        query_length = read.infer_read_length()

        if query_strand == '+':
            query_start = read.query_alignment_start + 1 
            query_end = read.query_alignment_end
        else:
            query_start = query_length - read.query_alignment_end + 1
            query_end = query_length - read.query_alignment_start

        inserted_pos = "---"
        for inserted_position in inserted_positions:
            ichr, istart, iend = inserted_position
            if reference_name == ichr and reference_end > istart - 5000 and reference_start < iend + 5000:
                inserted_pos = ichr + ',' + str(istart) + ',' + str(iend)

        print(query_name + '\t' + ','.join([str(query_start), str(query_end), query_strand, str(mapping_quality), reference_name, str(reference_start), str(reference_end)]) + '\t' + inserted_pos, file = hout)

    samfile.close()
    hout.close()



def summarize_bwa_alignment2(input_sam, seq_list, output_file):

    samfile = pysam.AlignmentFile(input_sam, 'r')


    inserted_positions = []
    with open(seq_list, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tchr, tstart, _, _, tend, _, _ = F[1].split(',')
            inserted_positions.append((tchr, int(tstart), int(tend)))


    samfile = pysam.AlignmentFile(input_sam, 'r')

    query2align_supp = {}
    query2align_primary = {}
    query2align_ins = {}

    hout = open(output_file, 'w') 
    for read in samfile.fetch():

        if read.is_unmapped: continue
        if read.is_secondary: continue

        query_name = read.query_name
        key = query_name

        query_strand = '-' if read.is_reverse else '+'
        reference_name = read.reference_name
        reference_start = read.reference_start + 1
        reference_end = read.reference_end
        mapping_quality = read.mapping_quality
        query_length = read.infer_read_length()

        cigartuples = read.cigartuples
        left_hard_clipping_size, right_hard_clipping_size = 0, 0
        total_ins_size, total_del_size = 0, 0
        if cigartuples[0][0] == 5: left_hard_clipping_size = cigartuples[0][1]
        if cigartuples[-1][0] == 5: right_hard_clipping_size = cigartuples[-1][1]

        for tcigar in cigartuples:
            if tcigar[0] == 1: total_ins_size = total_ins_size + tcigar[1]
            if tcigar[0] == 2: total_del_size = total_del_size + tcigar[1]

        total_mismatch_size = 0
        for ttag in read.get_tags():
            if ttag[0] == "NM": total_mismatch_size = total_mismatch_size + ttag[1]


        if not read.is_supplementary:
            if query_strand == '+':
                query_start = read.query_alignment_start + 1 
                query_end = read.query_alignment_end
            else:
                query_start = query_length - read.query_alignment_end + 1
                query_end = query_length - read.query_alignment_start
        else:
            if query_strand == '+':
                query_start = left_hard_clipping_size + 1
                query_end = query_length - right_hard_clipping_size
            else:
                query_start = right_hard_clipping_size + 1
                query_end = query_length - left_hard_clipping_size

        # align_info = ','.join([str(query_start), str(query_end), query_strand, str(mapping_quality), reference_name, str(reference_start), str(reference_end)])
        align_info = "%d,%d,%s,%d,%d,%d,%d,%s,%d,%d" % \
            (query_start, query_end, query_strand, mapping_quality, total_mismatch_size, total_ins_size, total_del_size, reference_name, reference_start, reference_end)

        if read.is_supplementary:
            if query_name not in query2align_supp: query2align_supp[query_name] = []
            query2align_supp[query_name].append(align_info + ",s")
        else:
            if read.mapping_quality >= 30:
                query2align_primary[query_name] = align_info + ",p"

                inserted_pos = "---"
                for inserted_position in inserted_positions:
                    ichr, istart, iend = inserted_position
                    if reference_name == ichr and reference_end > istart - 5000 and reference_start < iend + 5000:
                        inserted_pos = ichr + ',' + str(istart) + ',' + str(iend)
                query2align_ins[query_name] = inserted_pos

    samfile.close()

    hout = open(output_file, 'w')
    for query in sorted(query2align_primary):
        align_primary = query2align_primary[query]
        _, _, _, _, _, _, _, pchr, pstart, pend, _ = align_primary.split(',')
        align_ins = query2align_ins[query]
        align_info_final = align_primary
        if query in query2align_supp:
            for elm in query2align_supp[query]:
                _, _, _, _, _, _, _, schr, sstart, send, _ = elm.split(',')
                if int(pstart) - 5000 <= int(send) and int(pend) + 5000 >= int(sstart):
                    align_info_final = align_info_final + ';' + elm
        print("%s\t%s\t%s" % (query, align_info_final, align_ins), file = hout)


    hout.close()




def organize_info(rmsk_file, alignment_file, tsd_file, seq_list, output_file, LINE1_db_file):

    key2rmsk = {}
    with open(rmsk_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key2rmsk[F[0]] = [F[1], F[2], F[3], F[4], F[5]]

    key2alignment = {}
    with open(alignment_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key2alignment[F[0]] = [F[1], F[2]]


    key2tsd_polyAT = {}
    with open(tsd_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key2tsd_polyAT[F[0]] = [F[1], F[2]]

    keys = list(set(list(key2rmsk) + list(key2alignment) + list(key2tsd_polyAT)))

    """
    if genome_id == "hg38":
        line1_tb = pysam.TabixFile(pkg_resources.resource_filename("nanomonsv", "data/LINE1.hg38.bed.gz"))
    else:
        line1_tb = pysam.TabixFile(pkg_resources.resource_filename("nanomonsv", "data/LINE1.hg19.bed.gz"))
    """
    line1_tb = pysam.TabixFile(LINE1_db_file)

    with open(output_file, 'w') as hout:

        for key in keys:

            repeat_type, line1_ratio, alu_ratio, sva_ratio, rmsk_info = key2rmsk[key] if key in key2rmsk else ["None", 0.0, 0.0, 0.0, "---"]
            alignment_infos, inserted_pos = key2alignment[key] if key in key2alignment else ["---", "---"]
            is_polyAT, tsd = key2tsd_polyAT[key] if key in key2tsd_polyAT else ["---", "---"]
            
            line1_ratio = float(line1_ratio)
            line1_info = None
            overlap_ratio = None
            transduction_class = None

            if repeat_type.endswith("LINE1"): transduction_class = "Solo"

            if alignment_infos == "---" or is_polyAT == None: 
                print('\t'.join([key, repeat_type, str(line1_ratio), str(alu_ratio), str(sva_ratio), 
                                 rmsk_info, alignment_infos, inserted_pos, is_polyAT, tsd,
                                 str(line1_info), str(transduction_class)]), file = hout)
                continue


            achr, astart, aend, astrand = None, None, None, None
            aqstart, aqend = None, None
            for alignment_info in alignment_infos.split(';'):
                tqstart, tqend, tastrand, _, _, _, _, tachr, tastart, taend, _ = alignment_info.split(',')
                if achr is None:
                    achr, astart, aend, astrand, aqstart, aqend = tachr, int(tastart), int(taend), tastrand, int(tqstart), int(tqend)
                else:
                    if (is_polyAT == "polyT" and int(tqstart) < aqstart) or (is_polyAT == "polyA" and int(tqend) > aqend):
                        achr, astart, aend, astrand, aqstart, aqend = tachr, int(tastart), int(taend), tastrand, int(tqstart), int(tqend)

            if not achr.startswith("chr"): achr = "chr" + achr

            source_dir = '+' if (astrand == '+' and is_polyAT == "polyT") or (astrand == '-' and is_polyAT == "polyA") else '-'

            tabix_error_flag = 0
            if source_dir == '+':
                cur_L1_pos = float("Inf")
                is_rmsk, is_1000g, is_gnomad = False, False, False
                try:
                    records = line1_tb.fetch(achr, int(astart) - 50, int(aend) + 5000)
                except Exception as inst:
                    tabix_error_flag = 1
                if tabix_error_flag == 0:
                    for record in records:
                        FF = record.split('\t')
                        if FF[5] == '+': continue
                        if int(FF[1]) > cur_L1_pos: continue
                        line1_info = FF[3]

                        if "umary_LINE1" in line1_info:
                            if is_rmsk: continue
                            is_1000g = True
                        elif "gnomAD-SV" in line1_info:
                            if is_1000g or is_rmsk: continue
                            is_gnomad = True
                        else:
                            is_rmsk = True

                        if line1_ratio < 0.01:
                            transduction_class = "Orphan"
                        else:
                            transduction_class = "Partnered"
                        cur_L1_pos = int(FF[1])

            elif source_dir == '-':
                cur_L1_pos = -float("Inf")
                is_rmsk, is_1000g, is_gnomad = False, False, False
                try:
                    records = line1_tb.fetch(achr, int(astart) - 5000, int(aend) + 50)
                except Exception as inst:
                    tabix_error_flag = 1
                if tabix_error_flag == 0:
                    for record in records:
                        FF = record.split('\t')
                        if FF[5] == '-': continue
                        if int(FF[1]) < cur_L1_pos: continue 
                        line1_info = FF[3]

                        if "umary_LINE1" in line1_info:                  
                            if is_rmsk: continue
                            is_1000g = True
                        elif "gnomAD-SV" in line1_info:
                            if is_1000g or is_rmsk: continue
                            is_gnomad = True
                        else:
                            is_rmsk = True

                        if line1_ratio < 0.01:
                            transduction_class = "Orphan"
                        else:
                            transduction_class = "Partnered"
                        cur_L1_pos = int(FF[1])


            print('\t'.join([key, repeat_type, str(line1_ratio), str(alu_ratio), str(sva_ratio),
                             rmsk_info, alignment_infos, inserted_pos, is_polyAT, tsd, 
                             str(line1_info), str(transduction_class)]), file = hout)
 


def annotate_sv_file(sv_file, source_file, ppseudo_file, seq_list, output_file):

    sid2skey = {}
    with open(seq_list, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            sid2skey[F[0]] = F[1]

    skey2source_info = {}
    with open(source_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            skey = sid2skey[F[0]]
            skey2source_info[skey] = F[1:]

    skey2ppseudo_info = {}
    with open(ppseudo_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            skey = sid2skey[F[0]]
            if skey in skey2ppseudo_info:
                tmatch_ratio = float(skey2ppseudo_info[skey][1])
                if float(F[2]) > tmatch_ratio:
                    skey2ppseudo_info[skey] = F[1:4]
            else:
                skey2ppseudo_info[skey] = F[1:4]


    with open(sv_file, 'r') as hin, open(output_file, 'w') as hout:
        header = hin.readline().rstrip('\n')
        print(header + '\t' + '\t'.join(["Insert_Type", "Is_Inversion", "L1_Raito", "Alu_Ratio", "SV_Ratio", "RMSK_Info",
                                         "Alignment_Info", "Inserted_Pos", "Is_PolyA_T", "Target_Site_Duplication", "L1_Source_Info",
                                         "PSD_Gene", "PSD_Overlap_Ratio", "PDS_Exon_Num"]), file = hout) 
        for line in hin:
            F = line.rstrip('\n').split('\t')
            skey = ','.join(F[:6] + [str(len(F[6]))])

            if skey not in skey2source_info:
                print('\t'.join(F) + '\t' + "None" + '\t' + '\t'.join(["---"] * 13), file = hout)
                continue
        
            source_info = skey2source_info[skey]
            ppseudo_line = '\t'.join(skey2ppseudo_info[skey]) if skey in skey2ppseudo_info else "---\t---\t---"

            insert_type = "---"
            if source_info[10] == "Orphan":
                insert_type = "Orphan_L1"
            elif source_info[10] == "Partnered":
                insert_type = "Partnered_L1"
            elif source_info[10] == "Solo":
                insert_type = "Solo_L1"
            elif source_info[0] == "Alu":
                insert_type = "Alu"
            elif source_info[0] == "SVA":
                insert_type = "SVA"
            elif skey in skey2ppseudo_info:
                insert_type = "PSD"

            is_inversion = "NA"
            if source_info[0].startswith("Simple"):
                is_inversion = "Simple"
            elif source_info[0].startswith("Inverted"):
                is_inversion = "Inverted"
            elif source_info[0].startswith("Other"):
                is_inversion = "Other"

            print('\t'.join(F) + '\t' + insert_type + '\t' + is_inversion + '\t' + \
                  '\t'.join(source_info[1:10]) + '\t' + ppseudo_line, file = hout)






#! /usr/bin/env python3

import sys, csv, itertools
import pysam, parasail

from nanomonsv.my_seq import reverse_complement
from nanomonsv.logger import get_logger

logger = get_logger(__name__)

class Sv(object):

    def __init__(self, tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, sv_id, 
        total_read_tumor, var_read_tumor, total_read_ctrl, var_read_ctrl):

        self.chr1 = tchr1
        self.pos1 = tpos1
        self.dir1 = tdir1
        self.chr2 = tchr2
        self.pos2 = tpos2
        self.dir2 = tdir2
        self.inseq = tinseq
        self.sv_id = sv_id 
        self.total_read_tumor = total_read_tumor
        self.var_read_tumor = var_read_tumor
        self.total_read_ctrl = total_read_ctrl
        self.var_read_ctrl = var_read_ctrl
        self.filter = []


class Sv_filterer(object):

    def __init__(self, output_file, reference_fasta, is_control, simple_repeat_bed,  
        min_tumor_VAF = 0.05, min_indel_size = 50, bp_dist_margin = 30, validate_seg_len = 100):
        self.sv_list = []
        self.hout = open(output_file, 'w')
        self.bp_dist_margin = bp_dist_margin
        self.is_control = is_control
        self.reference_h = pysam.FastaFile(reference_fasta)
        self.min_tumor_VAF = min_tumor_VAF
        self.min_indel_size = min_indel_size
        self.bp_dist_margin = bp_dist_margin
        self.validate_seg_len = validate_seg_len
        self.simple_repeat_tb = pysam.TabixFile(simple_repeat_bed) if simple_repeat_bed is not None else None 
        self.header = None
        

    def __del__(self):
        self.hout.close()
        self.reference_h.close()


    def filter_close_both_breakpoints(self, sv1, sv2, filter_item = "Duplicate_with_close_SV"):

        # only apply when the first or the second sv is not insertion
        if len(sv1.inseq) >= self.min_indel_size or len(sv2.inseq) >= self.min_indel_size: return 

        if sv1.chr1 == sv2.chr1 and sv1.chr2 == sv2.chr2 and sv1.dir1 == sv2.dir1 and sv1.dir2 == sv2.dir2 and \
            abs(sv1.pos1 - sv2.pos1) < self.bp_dist_margin and abs(sv1.pos2 - sv2.pos2) < self.bp_dist_margin:

            if sv1.var_read_tumor is not None and sv2.var_read_tumor is not None:
                if sv1.var_read_tumor > sv2.var_read_tumor: 
                    sv1.filter.append(filter_item); return
                if sv2.var_read_tumor < sv1.var_read_tumor: 
                    sv2.filter.append(filter_item); return

            if len(sv1.inseq) < len(sv2.inseq):
                sv1.filter.append(filter_item); return
            if len(sv2.inseq) < len(sv1.inseq): 
                sv2.filter.append(filter_item); return

            if sv1.pos1 < sv2.pos1: 
                sv1.filter.append(filter_item); return
            if sv2.pos1 < sv1.pos1: 
                sv2.filter.append(filter_item); return

            if sv1.pos2 < sv2.pos2: 
                sv1.filter.append(filter_item); return
            if sv2.pos2 < sv1.pos2: 
                sv2.filter.append(filter_item) ; return

            sv2.filter.append(filter_item); return



    def filter_sv_insertion_match(self, sv, ins, filter_item = "Duplicate_with_insertion"):
    
        # only apply when the first sv is not insertion and the second sv is insertion type
        if len(sv.inseq) >= self.min_indel_size or len(ins.inseq) < self.min_indel_size: return 

        if not (sv.chr1 == ins.chr1 and abs(sv.pos1 - ins.pos1) <= self.bp_dist_margin) and \
            not (sv.chr2 == ins.chr2 and abs(sv.pos2 - ins.pos2) <= self.bp_dist_margin):
            return 

        ins_seg = self.reference_h.fetch(ins.chr1, max(ins.pos1 - self.validate_seg_len - self.bp_dist_margin - 1, 0), ins.pos1 - 1).upper()
        ins_seg = ins_seg + ins.inseq
        ins_seg = ins_seg + self.reference_h.fetch(ins.chr1, ins.pos2 - 1, ins.pos2 + self.validate_seg_len + self.bp_dist_margin - 1).upper()

        if sv.dir1 == '+':
            tseq = self.reference_h.fetch(sv.chr1, max(sv.pos1 - self.validate_seg_len - 1, 0), sv.pos1 - 1).upper()
        else:
            tseq = self.reference_h.fetch(sv.chr1, sv.pos1 - 1, sv.pos1 + self.validate_seg_len - 1).upper()
            tseq = reverse_complement(tseq)

        if sv.dir1 == '+':
            sv_seg = tseq + sv.inseq
        else:
            sv_seg = tseq + reverse_complement(sv.inseq)

        if sv.dir2 == '-':
            tseq = self.reference_h.fetch(sv.chr2, sv.pos2 - 1, sv.pos2 + self.validate_seg_len - 1).upper()
        else:
            tseq = self.reference_h.fetch(sv.chr2, max(sv.pos2 - self.validate_seg_len - 1, 0), sv.pos2 - 1).upper()
            tseq = reverse_complement(tseq)
 
        sv_seg = sv_seg + tseq

        user_matrix = parasail.matrix_create("ACGT", 2, -2)
        res = parasail.ssw(sv_seg, ins_seg, 3, 1, user_matrix)
        res_r = parasail.ssw(reverse_complement(sv_seg), ins_seg, 3, 1, user_matrix)

        if res.score1 > res_r.score1:
            match_ratio = float(res.score1) / (2 * (res.ref_end1 - res.ref_begin1 + 1))
            if match_ratio > 0.75:
                if res.read_begin1 < 0.1 * len(sv_seg) and res.read_end1 > 0.9 * len(sv_seg): 
                    sv.filter.append(filter_item)
                    return 
        else:
            match_ratio = float(res_r.score1) / (2 * (res_r.ref_end1 - res_r.ref_begin1 + 1))
            if match_ratio > 0.75:
                if res_r.read_begin1 < 0.1 * len(sv_seg) and res_r.read_end1 > 0.9 * len(sv_seg): 
                    sv.filter.append(filter_item)
                    return

        return


    def filter_dup_insertion(self, ins1, ins2, filter_item = "Duplicate_with_close_insertion"):

        # only apply when the first and the second sv is insertion type
        if len(ins1.inseq) < self.min_indel_size or len(ins2.inseq) < self.min_indel_size: return None

        bp_match = (ins1.chr1 == ins2.chr1 and abs(ins1.pos1 - ins2.pos1) <= 2 * self.bp_dist_margin) and \
            (ins1.chr2 == ins2.chr2 and abs(ins1.pos2 - ins2.pos2) <= 2 * self.bp_dist_margin)

        if bp_match == False: return None

        # insertion with shorter length will be filtered
        if len(ins1.inseq) < len(ins2.inseq):
            ins1.filter.append(filter_item)
            return
        else:
            ins2.filter.append(filter_item)


    def filter_sv_with_decoy(self, sv, filter_item = "SV_with_decoy"):

        if sv.chr1.endswith("decoy") or sv.chr2.endswith("decoy"): 
            sv.filter.append(filter_item)
            return


    def filter_small_sv(self, sv, filter_item = "Too_small_size"):

        if sv.chr1 == sv.chr2 and sv.dir1 == '+' and sv.dir2 == '-':
            sv_size = sv.pos2 - sv.pos1 + len(sv.inseq) - 1
            if sv_size < self.min_indel_size:
                sv.filter.append(filter_item)
                return

    def filter_low_vaf_sv(self, sv, filter_item = "Too_low_VAF"):

        if float(sv.var_read_tumor) / float(sv.total_read_tumor) < self.min_tumor_VAF:
            sv.filter.append(filter_item)

   
    def filter_simple_repeat(self, sv, filter_item = "Simple_repeat", simple_repeat_dist_margin = 30):

        if sv.chr1 == sv.chr2 and sv.dir1 == '+' and sv.dir2 == '-':
            sv_size = sv.pos2 - sv.pos1 + len(sv.inseq) - 1

            tabix_error_flag = False
            try:
                records = self.simple_repeat_tb.fetch(sv.chr1, max(sv.pos1 - simple_repeat_dist_margin + 1, 0), 
                    sv.pos1 + simple_repeat_dist_margin)
            except Exception as inst:
                print(f'{type(inst)}: {inst.args}', file = sys.stderr)
                tabix_error_flag = True

            if tabix_error_flag == False:
                for record_line in records:
                    record = record_line.split('\t')
                    if sv.pos1 >= int(record[1]) - simple_repeat_dist_margin and sv.pos2 <= int(record[2]) + simple_repeat_dist_margin:
                        sv.filter.append(filter_item)


            
    def add_sv(self, tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, sv_id,
        total_read_tumor, var_read_tumor, total_read_ctrl, var_read_ctrl):
    
        sv = Sv(tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, sv_id,
            total_read_tumor, var_read_tumor, total_read_ctrl, var_read_ctrl)
        self.sv_list.append(sv)


    def apply_filters(self):

        for sv in self.sv_list:
            self.filter_low_vaf_sv(sv)

        for sv in self.sv_list:
            self.filter_small_sv(sv)

        for sv in self.sv_list:
            self.filter_sv_with_decoy(sv)

        if self.simple_repeat_tb is not None:
            for sv in self.sv_list:
                self.filter_simple_repeat(sv)

        # logger.info("filter_close_both_breakpoints")
        for sv1, sv2 in itertools.combinations(self.sv_list, 2):
            if len(sv1.filter) > 0 or len(sv2.filter) > 0: continue
            self.filter_close_both_breakpoints(sv1, sv2)

        # logger.info("filter_sv_insertion_match")
        for sv1, sv2 in itertools.combinations(self.sv_list, 2):
            if len(sv1.filter) > 0 or len(sv2.filter) > 0: continue
            if len(sv1.inseq) < self.min_indel_size and len(sv2.inseq) >= self.min_indel_size: 
                self.filter_sv_insertion_match(sv1, sv2)
            elif len(sv2.inseq) < self.min_indel_size and len(sv1.inseq) >= self.min_indel_size:
                self.filter_sv_insertion_match(sv2, sv1)

        # logger.info("filter_dup_insertion")
        for ins1, ins2 in itertools.combinations(self.sv_list, 2):
            if len(ins1.filter) > 0 or len(ins2.filter) > 0: continue
            if len(ins1.inseq) >= self.min_indel_size and len(ins2.inseq) >= self.min_indel_size:
                self.filter_dup_insertion(ins1, ins2)


    def flush_sv_list(self):

        header = "Chr_1\tPos_1\tDir_1\tChr_2\tPos_2\tDir_2\tInserted_Seq\tSV_ID\tChecked_Read_Num_Tumor\tSupporting_Read_Num_Tumor" 
        if self.is_control: header = header + "\tChecked_Read_Num_Control\tSupporting_Read_Num_Control"
        header = header + '\t' + "Is_Filter"
        print(header, file = self.hout)
        for sv in self.sv_list:

            print_sv_line = f"{sv.chr1}\t{sv.pos1}\t{sv.dir1}\t{sv.chr2}\t{sv.pos2}\t{sv.dir2}\t{sv.inseq}\t{sv.sv_id}"
            print_sv_line = print_sv_line + f"\t{sv.total_read_tumor}\t{sv.var_read_tumor}"

            if self.is_control: print_sv_line = print_sv_line + f"\t{sv.total_read_ctrl}\t{sv.var_read_ctrl}"
            if len(sv.filter) == 0: 
                filter_print = "PASS"
            else:
                if "Duplicate_with_close_SV" in sv.filter: continue
                if "Duplicate_with_insertion" in sv.filter: continue
                if "Duplicate_with_close_insertion" in sv.filter: continue
                filter_print = ';'.join(list(set(sv.filter)))

            print(f"{print_sv_line}\t{filter_print}", file = self.hout)

    
def integrate_realignment_result(tumor_file, control_file, output_file, reference_fasta, simple_repeat_bed = None, min_indel_size = 50,
    min_tumor_variant_read_num = 3, min_tumor_VAF = 0.05, max_control_variant_read_num = 1, max_control_VAF = 0.03):

    is_control = True if control_file is not None else False
    svkey2control_info = {}
    if is_control:
        with open(control_file, 'r') as hin:
            for line in hin:
                F = line.rstrip('\n').split('\t')
                svkey = (F[0], int(F[1]), F[2], F[3], int(F[4]), F[5], F[6], F[7])
                svkey2control_info[svkey] = (int(F[8]), int(F[9]))

    sv_filterer = Sv_filterer(output_file, reference_fasta, is_control, simple_repeat_bed, min_tumor_VAF, min_indel_size)
    with open(tumor_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, tid = F[0], int(F[1]), F[2], F[3], int(F[4]), F[5], F[6], F[7]
            tkey = (tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, tid)

            total_read_tumor, var_read_tumor = int(F[8]), int(F[9])
            total_read_ctrl, var_read_ctrl = None, None

            if tkey in svkey2control_info: total_read_ctrl, var_read_ctrl = svkey2control_info[tkey][0], svkey2control_info[tkey][1]

            # filtering by supporting read and variant frequencies
            if total_read_tumor == 0: continue
            if var_read_tumor < min_tumor_variant_read_num: continue

            if is_control:
                if total_read_ctrl is None or total_read_ctrl == 0: continue
                if var_read_ctrl > max_control_variant_read_num: continue
                if float(var_read_ctrl) / float(total_read_ctrl) > max_control_VAF: continue


            sv_filterer.add_sv(tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, tid,
                total_read_tumor, var_read_tumor, total_read_ctrl, var_read_ctrl)

    sv_filterer.apply_filters()
    sv_filterer.flush_sv_list()
    del sv_filterer


def proc_sread_info_file(tumor_sread_info_file, sv_result_file, output_file, validate_sequence_length = 200):

    svkey = {}
    with open(sv_result_file, 'r') as hin:
        header = hin.readline()
        for line in hin:
            F = line.rstrip('\t').split('\t')
            svkey['\t'.join(F[:8])] = 1

    with open(tumor_sread_info_file, 'r') as hin, open(output_file, 'w') as hout:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tkey = '\t'.join(F[:8])
            if tkey not in svkey: continue

            score1, cstart1, cend1, sstart1, send1, strand1 = int(F[9]), int(F[10]), int(F[11]), int(F[12]), int(F[13]), F[14]
            score2, cstart2, cend2, sstart2, send2, strand2 = int(F[15]), int(F[16]), int(F[17]), int(F[18]), int(F[19]), F[20]

            tinseq = F[6]
            readid = F[8]
                
            if score1 >= score2:
                sstrand = strand1
                if sstrand == '+':
                    spos = (float(send1 - sstart1) / float(cend1 - cstart1)) * (validate_sequence_length - cstart1) + sstart1
                else:
                    spos = (float(sstart1 - send1) / float(cend1 - cstart1)) * (validate_sequence_length - cstart1) + send1
            else:
                sstrand = strand2
                if sstrand == '+':
                    spos = (float(send2 - sstart2) / float(cend2 - cstart2)) * (validate_sequence_length - cstart2) + sstart2 - len(tinseq) - 1
                else:
                    spos = (float(sstart2 - send2) / float(cend2 - cstart2)) * (validate_sequence_length - cstart2) + send2 + len(tinseq) + 1
            
            print(f"{tkey}\t{readid}\t{int(spos)}\t{sstrand}", file = hout)


def filter_sbnd_in_simple_repeat(tchr, tpos, simple_repeat_tb, margin=50):
    tabix_error_flag = False
    try:
        records = simple_repeat_tb.fetch(tchr, max(tpos - margin, 0), tpos + margin)
    except Exception as inst:
        print(f'{type(inst)}: {inst.args}', file=sys.stderr)
        tabix_error_flag = True

    if not tabix_error_flag:
        for record_line in records:
            record = record_line.split('\t')
            if tpos >= int(record[1]) - margin and tpos <= int(record[2]) + margin:
                return True
    return False


def integrate_realignment_result_sbnd(tumor_sbnd_count_file, ctrl_sbnd_count_file, output_file,
    nonsbnd_result_file, refined_bp_sbnd_file,
    min_tumor_variant_read_num = 3, min_tumor_VAF = 0.05, max_control_variant_read_num = 1, max_control_VAF = 0.03,
    simple_repeat_bed = None):

    margin_bp = 50
    nanomonsv_bp_list = []
    with open(nonsbnd_result_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            if F[0] == "Chr_1": continue
            nanomonsv_bp_list.append((F[0], int(F[1]), F[2]))
            nanomonsv_bp_list.append((F[3], int(F[4]), F[5]))

    key2count_ctrl = {}
    if ctrl_sbnd_count_file is not None:
        with open(ctrl_sbnd_count_file, 'r') as hin:
            for line in hin:
                F = line.rstrip('\n').split('\t')
                key = F[4]
                key2count_ctrl[key] = (F[5], F[6])

    key2contig = {}
    with open(refined_bp_sbnd_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key = F[4]
            key2contig[key] = F[3]

    simple_repeat_tb = None
    if simple_repeat_bed is not None:
        simple_repeat_tb = pysam.TabixFile(simple_repeat_bed)

    with open(tumor_sbnd_count_file, 'r') as hin, open(output_file, 'w') as hout:
        header = "Chr_1\tPos_1\tDir_1\tContig\tSV_ID\tChecked_Read_Num_Tumor\tSupporting_Read_Num_Tumor"
        if ctrl_sbnd_count_file is not None: header = header + "\tChecked_Read_Num_Control\tSupporting_Read_Num_Control"
        header = header + "\tIs_Filter"
        print(header, file = hout)

        for line in hin:
            F = line.rstrip('\n').split('\t')
            key = F[4]
            canonical_overlap = False
            for bp in nanomonsv_bp_list:
                if F[0] == bp[0] and F[2] == bp[2] and abs(int(F[1]) - bp[1]) < margin_bp:
                    canonical_overlap = True
                    break

            if ctrl_sbnd_count_file is not None:
                if key not in key2count_ctrl:
                    continue
                else:
                    ctrl_count = key2count_ctrl[key]

                if int(ctrl_count[0]) == 0: continue
                if int(ctrl_count[1]) > max_control_variant_read_num: continue
                if float(ctrl_count[1]) / float(ctrl_count[0]) > max_control_VAF: continue

            if int(F[6]) < min_tumor_variant_read_num: continue
            if float(F[6]) / float(F[5]) < min_tumor_VAF: continue

            filter_list = []
            if simple_repeat_tb is not None and filter_sbnd_in_simple_repeat(F[0], int(F[1]), simple_repeat_tb):
                filter_list.append("Simple_repeat")
            if canonical_overlap:
                filter_list.append("Canonical_SV_overlap")
            is_filter = ";".join(filter_list) if filter_list else "PASS"

            if ctrl_sbnd_count_file is not None:
                print(f"{F[0]}\t{F[1]}\t{F[2]}\t{key2contig[key]}\t{F[4]}\t{F[5]}\t{F[6]}\t{ctrl_count[0]}\t{ctrl_count[1]}\t{is_filter}", file = hout)
            else:
                print(f"{F[0]}\t{F[1]}\t{F[2]}\t{key2contig[key]}\t{F[4]}\t{F[5]}\t{F[6]}\t{is_filter}", file = hout)
    

if __name__ == "__main__":

    import sys
    integrate_realignment_result(sys.argv[1], sys.argv[2], sys.argv[3])




#! /usr/bin/env python3

import sys, os
from subprocess import Popen, PIPE
import pysam

from .logger import get_logger
logger = get_logger(__name__)

def is_exists_bam(input_file):

    if input_file.startswith("s3://"):
        is_exists_s3(input_file)
    else:
        is_exists(input_file)


def is_exists(input_file):
    
    if not os.path.exists(input_file):
        logger.error("Input not exists: %s" % input_file)
        sys.exit(1)


def is_exists_s3(bam_object):

    from urllib.parse import urlparse
    import boto3

    obj_p = urlparse(bam_object)

    tbucket = obj_p.netloc
    tkey = obj_p.path
    if tkey.startswith("/"): tkey = tkey[1:]

    client = boto3.client("s3")
    try:
        response = client.head_object(Bucket = tbucket, Key = tkey)
    except:
        logger.error("Input not exists: %s: " % bam_object)
        sys.exit(1)


def is_exists_parsed_files(input_prefix):

    missing_files = []
    if not os.path.exists(input_prefix + ".deletion.sorted.bed.gz"):
        missing_files.append(input_prefix + ".deletion.sorted.bed.gz")
    if not os.path.exists(input_prefix + ".deletion.sorted.bed.gz.tbi"):
        missing_files.append(input_prefix + ".deletion.sorted.bed.gz.tbi")
    if not os.path.exists(input_prefix + ".insertion.sorted.bed.gz"):
        missing_files.append(input_prefix + ".insertion.sorted.bed.gz")
    if not os.path.exists(input_prefix + ".insertion.sorted.bed.gz.tbi"):
        missing_files.append(input_prefix + ".insertion.sorted.bed.gz.tbi")
    if not os.path.exists(input_prefix + ".rearrangement.sorted.bedpe.gz"):
        missing_files.append(input_prefix + ".rearrangement.sorted.bedpe.gz")
    if not os.path.exists(input_prefix + ".rearrangement.sorted.bedpe.gz.tbi"):
        missing_files.append(input_prefix + ".rearrangement.sorted.bedpe.gz.tbi")

    if len(missing_files) > 0:
        logger.error("One or more files via the nanomonsv parse stage do not exist. " + \
                     "Please check the prefix path. Also confirm whether nanomonsv parse ended properly. " +  \
                     "Missing files: " + ' '.join(missing_files))
        sys.exit(1)


def get_alignment_object(alignment_file, reference_fasta):

    _, file_extension = os.path.splitext(alignment_file)
    file_extension = file_extension.lower()

    if file_extension == ".bam":
        return pysam.AlignmentFile(alignment_file, "rb")
    elif file_extension == ".cram":
        return pysam.AlignmentFile(alignment_file, "rc", reference_filename = reference_fasta)
    else:
        raise ValueError(f"Invalid file extension: {file_extension}. Expected .bam or .cram")


def bam_cram_format_check(alignment_file, reference_fasta):

    try:
        bam_t = get_alignment_object(alignment_file, reference_fasta)
    except:
        logger.error("Alignment file (BAM or CRAM) format error: %s" % alignment_file)
        sys.exit(1)

def fasta_format_check(fasta_file):

    try:
        fasta_t = pysam.FastaFile(fasta_file)
    except:
        logger.error("FASTA format error: %s" % fasta_file)
        sys.exit(1)
       
 
def is_tool(executable):

    from shutil import which
    if which(executable) is None:
        logger.error("Executable does not exist: " + executable)
        sys.exit(1) 

    return True


def libssw_check():

    # check whether libssw.so is in LD_LIBRARY_PATH
    sLibPath = ""
    for ld_path in os.environ["LD_LIBRARY_PATH"].split(':'):
        # print ld_path
        if os.path.exists(ld_path + "/libssw.so"):
            sLibPath = ld_path # + "/libssw.so"
            break
    if sLibPath == "":
        logger.error("Cannot find libssw.so in LD_LIBRARY_PATH")
        sys.exit(1)




#!/usr/bin/env python3
"""
Simple python wrapper for SSW library
Please put the path of libssw.so into LD_LIBRARY_PATH or pass it explicitly as a parameter
By Yongan Zhao (March 2016)
"""

from __future__ import print_function

import sys
import os.path as op
import argparse as ap
import ctypes as ct
import timeit as ti
import gzip
import math

from . import ssw_lib

if sys.version_info.major == 2:
    range = xrange



def read(sFile):
    """
    read a sequence file
    @param  sFile   sequence file
    """
    def read_one_fasta(f):
        """
        read a fasta file
        @param  f   file handler
        """
        sId = ''
        sSeq = ''
        for l in f:
            if l.startswith('>'):
                if sSeq:
                    yield sId, sSeq, ''
                sId = l.strip()[1:].split()[0]
                sSeq = ''
            else:
                sSeq += l.strip()

        yield sId, sSeq, ''

    def read_one_fastq(f):
        """
        read a fastq file
        @param  f   file handler
        """
        sId = ''
        sSeq = ''
        s3 = ''
        sQual = ''
        """
        for l in f:
            sId = l.strip()[1:].split()[0]
            sSeq = f.next()
            s3 = f.next()
            sQual = f.next()

            yield sId, sSeq, sQual
        """

        sID = f.readline()
        while sId:
            sId = sId.rstrip('\n').split()[0]
            sSeq = f.readline().rstrip('\n')
            sId2 = f.readline().rstrip('\n')
            sQual = f.readline().rstrip('\n')
            yield sId, sSeq, sQual
            sID = f.readline()


        f.close()

# test if fasta or fastq
    bFasta = True
    ext = op.splitext(sFile)[1][1:].strip().lower()
    if ext == 'gz' or ext == 'gzip':
        with gzip.open(sFile, 'r') as f:
            # l = f.next()
            l = f.readline().rstrip('\n')
            if l.startswith('>'):
                bFasta = True
            elif l.startswith('@'):
                bFasta = False
            else:
                print('file format cannot be recognized', file = sys.stderr)
                sys.exit()
    else:
        with open(sFile, 'r') as f:
            # l = f.next()
            l = f.readline().rstrip('\n')
            if l.startswith('>'):
                bFasta = True
            elif l.startswith('@'):
                bFasta = False
            else:
                print('file format cannot be recognized', file = sys.stderr)
                sys.exit()

# read
    if ext == 'gz' or ext == 'gzip':
        with gzip.open(sFile, 'r') as f:
            if bFasta == True:
                for sId,sSeq,sQual in read_one_fasta(f):
                    yield sId, sSeq, sQual
            else:
                for sId,sSeq,sQual in read_one_fastq(f):
                    yield sId, sSeq, sQual
    else:
        with open(sFile, 'r') as f:
            if bFasta == True:
                for sId,sSeq,sQual in read_one_fasta(f):
                    yield sId, sSeq, sQual
            else:
                for sId,sSeq,sQual in read_one_fastq(f):
                    yield sId, sSeq, sQual


def to_int(seq, lEle, dEle2Int):
    """
    translate a sequence into numbers
    @param  seq   a sequence
    """
    num_decl = len(seq) * ct.c_int8
    num = num_decl()
    for i,ele in enumerate(seq):
        try:
            n = dEle2Int[ele]
        except KeyError:
            n = dEle2Int[lEle[-1]]
        finally:
            num[i] = n

    return num


def align_one(ssw, qProfile, rNum, nRLen, nOpen, nExt, nFlag, nMaskLen):
    """
    align one pair of sequences
    @param  qProfile   query profile
    @param  rNum   number array for reference
    @param  nRLen   length of reference sequence
    @param  nFlag   alignment flag
    @param  nMaskLen   mask length
    """
    res = ssw.ssw_align(qProfile, rNum, ct.c_int32(nRLen), nOpen, nExt, nFlag, 0, 0, nMaskLen)

    nScore = res.contents.nScore
    nScore2 = res.contents.nScore2
    nRefBeg = res.contents.nRefBeg
    nRefEnd = res.contents.nRefEnd
    nQryBeg = res.contents.nQryBeg
    nQryEnd = res.contents.nQryEnd
    nRefEnd2 = res.contents.nRefEnd2
    lCigar = [res.contents.sCigar[idx] for idx in range(res.contents.nCigarLen)]
    nCigarLen = res.contents.nCigarLen
    ssw.align_destroy(res)

    return (nScore, nScore2, nRefBeg, nRefEnd, nQryBeg, nQryEnd, nRefEnd2, nCigarLen, lCigar)


def buildPath(q, r, nQryBeg, nRefBeg, lCigar):
    """
    build cigar string and align path based on cigar array returned by ssw_align
    @param  q   query sequence
    @param  r   reference sequence
    @param  nQryBeg   begin position of query sequence
    @param  nRefBeg   begin position of reference sequence
    @param  lCigar   cigar array
    """
    sCigarInfo = 'MIDNSHP=X'
    sCigar = ''
    sQ = ''
    sA = ''
    sR = ''
    nQOff = nQryBeg
    nROff = nRefBeg
    for x in lCigar:
        n = x >> 4
        m = x & 15
        if m > 8:
            c = 'M'
        else:
            c = sCigarInfo[m]
        sCigar += str(n) + c

        if c == 'M':
            sQ += q[nQOff : nQOff+n]
            sA += ''.join(['|' if q[nQOff+j] == r[nROff+j] else '*' for j in range(n)])
            sR += r[nROff : nROff+n]
            nQOff += n
            nROff += n
        elif c == 'I':
            sQ += q[nQOff : nQOff+n]
            sA += ' ' * n
            sR += '-' * n
            nQOff += n
        elif c == 'D':
            sQ += '-' * n
            sA += ' ' * n
            sR += r[nROff : nROff+n]
            nROff += n

    return sCigar, sQ, sA, sR




def main(args):
    lEle = []
    dRc = {} 
    dEle2Int = {}
    dInt2Ele = {}
    if False == args.bProtien:
# init DNA score matrix
        if not args.sMatrix:
            lEle = ['A', 'C', 'G', 'T', 'N']
            dRc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'a':'T', 'c':'G', 'g':'C', 't':'A'} 
            for i,ele in enumerate(lEle):
                dEle2Int[ele] = i
                dEle2Int[ele.lower()] = i
                dInt2Ele[i] = ele
            nEleNum = len(lEle)
            lScore = [0 for i in range(nEleNum**2)]
            for i in range(nEleNum-1):
                for j in range(nEleNum-1):
                    if lEle[i] == lEle[j]:
                        lScore[i*nEleNum+j] = args.nMatch
                    else:
                        lScore[i*nEleNum+j] = -args.nMismatch
        else:
            lEle, dEle2Int, dInt2Ele, lScore = ssw.read_matrix(args.sMatrix)
    else:
# load AA score matrix
        if not args.sMatrix:
            lEle = 'A   R   N   D   C   Q   E   G   H   I   L   K   M   F   P   S   T   W   Y   V   B   Z   X   *'.split()
            for i,ele in enumerate(lEle):
                dEle2Int[ele] = i
                dEle2Int[ele.lower()] = i
                dInt2Ele[i] = ele
            nEleNum = len(lEle)
            lScore = ssw_lib.lBlosum50
        else:
            # assume the format of the input score matrix is the same as that of http://www.ncbi.nlm.nih.gov/Class/FieldGuide/BLOSUM62.txt
            lEle, dEle2Int, dInt2Ele, lScore = ssw.read_matrix(args.sMatrix)

    if args.bBest and args.bProtien:
        print('Reverse complement alignment is not available for protein sequences.', file = sys.stderr)

# translate score matrix to ctypes
    mat = (len(lScore) * ct.c_int8) ()
    mat[:] = lScore
# set flag
    nFlag = 0
    if args.bPath:
        nFlag = 2
# print sam head
    if args.bSam and args.bHeader and args.bPath:
        print('@HD\tVN:1.4\tSO:queryname')
        for sRId,sRSeq,_ in read(args.target):
            print('@SQ\tSN:{}\tLN:{}'.format(sRId, len(sRSeq)))
    elif args.bSam and not args.bPath:
        print('SAM format output is only available together with option -c.\n', file = sys.stderr)
        args.bSam = False

    ssw = ssw_lib.CSsw(args.sLibPath)
# iterate query sequence
    for sQId,sQSeq,sQQual in read(args.query):
# build query profile
        qNum = to_int(sQSeq, lEle, dEle2Int)
        qProfile = ssw.ssw_init(qNum, ct.c_int32(len(sQSeq)), mat, len(lEle), 2)
# build rc query profile
        if args.bBest and not args.bProtien:
            sQRcSeq = ''.join([dRc[x] for x in sQSeq[::-1]])
            qRcNum = to_int(sQRcSeq, lEle, dEle2Int)
            qRcProfile = ssw.ssw_init(qRcNum, ct.c_int32(len(sQSeq)), mat, len(lEle), 2)
# set mask len
        if len(sQSeq) > 30:
            nMaskLen = int(len(sQSeq) / 2)
        else:
            nMaskLen = 15

# iter target sequence
        for sRId,sRSeq,_ in read(args.target):
            rNum = to_int(sRSeq, lEle, dEle2Int)

# format ofres: (nScore, nScore2, nRefBeg, nRefEnd, nQryBeg, nQryEnd, nRefEnd2, nCigarLen, lCigar)
            res = align_one(ssw, qProfile, rNum, len(sRSeq), args.nOpen, args.nExt, nFlag, nMaskLen)
# align rc query
            resRc = None
            if args.bBest and not args.bProtien:
                resRc = align_one(ssw, qRcProfile, rNum, len(sRSeq), args.nOpen, args.nExt, nFlag, nMaskLen)

# build cigar and trace back path
            strand = 0
            if resRc == None or res[0] > resRc[0]:
                resPrint = res
                strand = 0
                sCigar, sQ, sA, sR = buildPath(sQSeq, sRSeq, res[4], res[2], res[8])
            else:
                resPrint = resRc
                strand = 1
                sCigar, sQ, sA, sR = buildPath(sQRcSeq, sRSeq, resRc[4], resRc[2], resRc[8])

# print results
            if not args.bSam:
                print('target_name: {}\nquery_name: {}\noptimal_alignment_score: {}\t'.format(sRId, sQId, resPrint[0]), end = '')
                if resPrint[1] > 0:
                    print('suboptimal_alignment_score: {}\t'.format(resPrint[1]), end = '')
                if strand == 0:
                    print('strand: +\t', end = '')
                else: 
                    print('strand: -\t', end = '')
                if resPrint[2] + 1:
                    print('target_begin: {}\t'.format(resPrint[2] + 1), end = '')
                print('target_end: {}\t'.format(resPrint[3] + 1), end = '')
                if resPrint[4] + 1:
                    print('query_begin: {}\t'.format(resPrint[4] + 1), end = '')
                print('query_end: {}\n'.format(resPrint[5] + 1))
                if resPrint[-2] > 0:
                    n1 = 1 + resPrint[2]
                    n2 = min(60,len(sR)) + resPrint[2] - sR.count('-',0,60)
                    n3 = 1 + resPrint[4]
                    n4 = min(60,len(sQ)) + resPrint[4] - sQ.count('-',0,60)
                    for i in range(0, len(sQ), 60):
                        print('Target:{:>8}\t{}\t{}'.format(n1, sR[i:i+60], n2))
                        n1 = n2 + 1
                        n2 = n2 + min(60,len(sR)-i-60) - sR.count('-',i+60,i+120)

                        print('{: ^15}\t{}'.format('', sA[i:i+60]))

                        print('Query:{:>9}\t{}\t{}\n'.format(n3, sQ[i:i+60], n4))
                        n3 = n4 + 1
                        n4 = n4 + min(60,len(sQ)-i-60) - sQ.count('-',i+60,i+120)
            else:
                print("{}\t".format(sQId), end = '')
                if resPrint[0] == 0:
                    print("4\t*\t0\t255\t*\t*\t0\t0\t*\t*", end = '')
                else:
                    mapq = 1-abs(resPrint[0]-resPrint[1])/float(resPrint[0]) 
                    mapq = -4.343 * math.log(mapq) if mapq > 1e-16 else -4.343 * math.log(1e-16)
                    # mapq = int(-4.343 * math.log(1-abs(resPrint[0]-resPrint[1])/float(resPrint[0])))
                    mapq = int(mapq + 4.99);
                    if mapq >= 254:
                        mapq = 254
                    if strand == 1:
                        print('16\t', end = '')
                    else:
                        print('0\t', end = '')
                    print('{}\t{}\t{}\t'.format(sRId, resPrint[2]+1, mapq), end = '')
                    print(sCigar, end = '')
                    print('\t*\t0\t0\t', end = '')
                    print(sQSeq[resPrint[4]:resPrint[5]+1] if strand==0 else sQRcSeq[resPrint[4]:resPrint[5]+1], end = '')
                    print('\t', end = '')
                    if sQQual:
                        if strand == 0:
                            print(sQQual[resPrint[4]:resPrint[5]+1], end = '')
                        else:
                            print(sQQual[-resPrint[4]-1:-resPrint[5]-1:-1])
                    else:
                        print('*', end = '')

                    print('\tAS:i:{}'.format(resPrint[0]), end = '')
                    print('\tNM:i:{}\t'.format(len(sA)-sA.count('|')), end = '')
                    if resPrint[1] > 0:
                        print('ZS:i:{}'.format(resPrint[1]))
                    else:
                        print()


        ssw.init_destroy(qProfile)
        if args.bBest and not args.bProtien:
            ssw.init_destroy(qRcProfile)


if __name__ == '__main__':

    parser = ap.ArgumentParser()
    parser.add_argument('-l', '--sLibPath', default='', help='path of libssw.so')
    parser.add_argument('-m', '--nMatch', type=int, default=2, help='a positive integer as the score for a match in genome sequence alignment. [default: 2]')
    parser.add_argument('-x', '--nMismatch', type=int, default=2, help='a positive integer as the score for a mismatch in genome sequence alignment. [default: 2]')
    parser.add_argument('-o', '--nOpen', type=int, default=3, help='a positive integer as the penalty for the gap opening in genome sequence alignment. [default: 3]')
    parser.add_argument('-e', '--nExt', type=int, default=1, help='a positive integer as the penalty for the gap extension in genome sequence alignment. [default: 1]')
    parser.add_argument('-p', '--bProtien', action='store_true', help='Do protein sequence alignment. Without this option, the ssw_test will do genome sequence alignment. [default: False]')
    parser.add_argument('-a', '--sMatrix', default='', help='a file for either Blosum or Pam weight matrix. [default: Blosum50]')
    parser.add_argument('-c', '--bPath', action='store_true', help='Return the alignment path. [default: False]')
    parser.add_argument('-f', '--nThr', default=0, help='a positive integer. Only output the alignments with the Smith-Waterman score >= N.')
    parser.add_argument('-r', '--bBest', action='store_true', help='The best alignment will be picked between the original read alignment and the reverse complement read alignment. [default: False]')
    parser.add_argument('-s', '--bSam', action='store_true', help='Output in SAM format. [default: no header]')
    parser.add_argument('-header', '--bHeader', action='store_true', help='If -s is used, include header in SAM output.')
    parser.add_argument('target', help='targe file')
    parser.add_argument('query', help='query file')
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()
    args = parser.parse_args()

    t1 = ti.default_timer()
    main(args)
    t2 = ti.default_timer()
    print('CPU time: {} seconds'.format(t2 - t1), file = sys.stderr)




#! /usr/bin/env python3

import sys, os, subprocess, statistics 
import pysam

from .logger import get_logger
from .my_seq import reverse_complement
from .utils import get_alignment_object

logger = get_logger(__name__)

 
def set_readid2alignment(readid2alignment, input_file, mode, alignment_margin):

    # readid2alignment = {}
    cid = 0
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key = f"{F[0]},{F[1]},{F[2]},{F[8]},{F[3]},{F[4]},{F[5]},{F[9]},{mode},{cid}"
                # ','.join([F[0], F[1], F[2], F[8], F[3], F[4], F[5], F[9], mode])
            # readids = F[6].split(';')
            readids = eval(F[6])

            if mode == "r":
                info1 = F[10].split(';')
                info2 = F[11].split(';')
                for i in range(len(readids)):
                    tinfo1 = info1[i].split(',') 
                    tinfo2 = info2[i].split(',')
                    start1, end1, start2, end2 = int(tinfo1[0]), int(tinfo1[2]), int(tinfo2[0]), int(tinfo2[2])
                    if readids[i] not in readid2alignment: readid2alignment[readids[i]] = []
                    if start1 <= start2:
                        readid2alignment[readids[i]].append((key, max(end1 - alignment_margin, 1), start2 + alignment_margin, '+', '*'))
                    else:
                        readid2alignment[readids[i]].append((key, max(end2 - alignment_margin, 1), start1 + alignment_margin, '-', '*'))

            elif mode == "d":
                size = F[10].split(';')
                info = F[11].split(';')
                for i in range(len(readids)):
                    tinfo = info[i].split(',')
                    tpos, tlen, tstrand = int(tinfo[1]), int(tinfo[3]), tinfo[4]
                    if readids[i] not in readid2alignment: readid2alignment[readids[i]] = []
                    readid2alignment[readids[i]].append((key, max(tpos - alignment_margin, 1), min(tpos + alignment_margin, tlen - 1), tstrand, size[i]))

            elif mode == "i":
                size = F[10].split(';')
                info = F[11].split(';')
                median_size = round(statistics.median([int(x) for x in size if x not in ['-', '+']]))
                for i in range(len(readids)):
                    tinfo = info[i].split(',')
                    tpos, tlen, tstrand = int(tinfo[1]), int(tinfo[3]), tinfo[4]
                    if readids[i] not in readid2alignment: readid2alignment[readids[i]] = []
                    if size[i] in ['-', '+']: # breakpoint
                        if (tstrand == '+' and size[i] == '+') or (tstrand == '-' and size[i] == '-'):
                            readid2alignment[readids[i]].append((key, max(tpos - alignment_margin, 1), min(tpos + median_size + alignment_margin, tlen - 1), tstrand, size[i]))
                        else:
                            readid2alignment[readids[i]].append((key, max(tpos - median_size - alignment_margin, 1), min(tpos + alignment_margin, tlen - 1), tstrand, size[i]))
                    else: # Insertion identified by CIGAR
                        if tstrand == '+':
                            readid2alignment[readids[i]].append((key, max(tpos - alignment_margin, 1), min(tpos + int(size[i]) + alignment_margin, tlen - 1), tstrand, size[i]))
                        else:
                            readid2alignment[readids[i]].append((key, max(tpos - int(size[i]) - alignment_margin, 1), min(tpos + alignment_margin, tlen - 1), tstrand, size[i]))

            cid = cid + 1



def set_readid2alignment_sbnd(readid2alignment_sbnd, input_file, alignment_margin):

    cid = 0
    with open(input_file, 'r') as hin:
        for line in hin:
            tchr, tstart, tend, treadids, _, tstrand, tinfos = line.rstrip('\n').split('\t')
            tkey = f"{tchr},{tstart},{tend},{tstrand},b,{cid}" # ','.join([tchr, tstart, tend, tstrand])
           
            # readids = treadids.split(';')
            readids = eval(treadids)

            infos = tinfos.split(';') 
            for i in range(len(readids)):
                ttinfo = infos[i].split(',')
                qpos, qlen, qstrand = int(ttinfo[1]), int(ttinfo[3]), ttinfo[4]             
                if readids[i] not in readid2alignment_sbnd: readid2alignment_sbnd[readids[i]] = []
                if tstrand == qstrand: 
                    readid2alignment_sbnd[readids[i]].append((tkey, max(1, qpos - alignment_margin), qlen, qstrand, '*'))
                else:
                    readid2alignment_sbnd[readids[i]].append((tkey, 1, min(qpos + alignment_margin, qlen), qstrand, '*'))

            cid = cid + 1


def gather_support_read_seq(rearrangement_file, insertion_file, deletion_file, output_file, tumor_alignment_file, reference_fasta, 
    single_breakend_file = None, output_file_sbind = None, alignment_margin = 300):

    is_sbnd = True if single_breakend_file is not None else False

    
    alignment_h = get_alignment_object(tumor_alignment_file, reference_fasta)

    readid2alignment = {}
    set_readid2alignment(readid2alignment, rearrangement_file, 'r', alignment_margin)
    set_readid2alignment(readid2alignment, insertion_file, 'i', alignment_margin)
    set_readid2alignment(readid2alignment, deletion_file, 'd', alignment_margin)
    
    hout = open(output_file + ".tmp.unsorted", 'w')

    if is_sbnd:
        readid2alignment_sbnd = {}
        set_readid2alignment_sbnd(readid2alignment_sbnd, single_breakend_file, alignment_margin)
        hout_sbnd = open(output_file + ".sbnd.tmp.unsorted", 'w')

    for read in alignment_h.fetch():

        if read.is_secondary or read.is_supplementary: continue

        if read.query_name in readid2alignment:

            for talignment in readid2alignment[read.query_name]:

                akey, astart, aend, astrand, asize = talignment

                read_seq = reverse_complement(read.query_sequence) if read.is_reverse else read.query_sequence
                part_seq = read_seq[(astart - 1):aend]
                if astrand == '-': part_seq = reverse_complement(part_seq)

                print('\t'.join([akey, read.query_name, asize, part_seq]), file = hout)


        if is_sbnd and read.query_name in readid2alignment_sbnd: 
        
            for talignment in readid2alignment_sbnd[read.query_name]:

                akey, astart, aend, astrand, asize = talignment
                tchr, tstart, tend, tstrand, _, tcid = akey.split(',')

                read_seq = reverse_complement(read.query_sequence) if read.is_reverse else read.query_sequence
                part_seq = read_seq[(astart - 1):aend]
                if astrand != tstrand: part_seq = reverse_complement(part_seq)

                print('\t'.join([akey, read.query_name, asize, part_seq]), file = hout_sbnd)


    alignment_h.close()
    hout.close()
    if is_sbnd: hout_sbnd.close()

    with open(output_file, 'w') as hout:
        subprocess.check_call(["sort", "-k1,1", output_file + ".tmp.unsorted"], stdout = hout)
    os.remove(output_file + ".tmp.unsorted")

    if is_sbnd:
        with open(output_file_sbind, 'w') as hout:
            subprocess.check_call(["sort", "-k1,1", output_file + ".sbnd.tmp.unsorted"], stdout = hout)
        os.remove(output_file + ".sbnd.tmp.unsorted")





#! /usr/bin/env python3

import unittest
import subprocess

class TestMain(unittest.TestCase):

    def setUp(self):
        pass

    def test_command(self):
        subprocess.run(["nanomonsv", "--version"])
        

if __name__ == "__main__":
    unittest.main()




#! /usr/bin/env python3

import sys, os, subprocess, shutil, statistics 
import pysam
import parasail

from .my_seq import reverse_complement

from .logger import get_logger

logger = get_logger(__name__)


def get_refined_bp_sbnd(tconsensus, fasta_file_h, tchr, tstart, tend, tdir, hout_log, margin = 200):

    tconsensus_part = tconsensus[:1000] if len(tconsensus) > 1000 else tconsensus

    ref_len = fasta_file_h.get_reference_length(tchr)
    if tstart < 1: tstart = 1
    if ref_len < tend: tend = ref_len 

    if tdir == '+':
        qseq = fasta_file_h.fetch(tchr, max(int(tstart) - margin, 0), int(tend)).upper()
    else:
        qseq = fasta_file_h.fetch(tchr, max(int(tstart) - 1, 0), int(tend) + margin).upper()
        qseq = reverse_complement(qseq)

    user_matrix = parasail.matrix_create("ACGT", 1, -2)
    res = parasail.ssw(qseq, tconsensus, 3, 1, user_matrix)
    if res is None:
        logger.debug(f"Alignment for breakpoint localization failed for {tchr},{tstart},{tend},{tdir}")
        return None


    if tdir == '+':
        bp_pos_reference = tend - (len(qseq) - res.read_end1 - 1)
    else:
        bp_pos_reference = tstart + (len(qseq) - res.read_end1 - 1) 

    tconsensus_after = tconsensus[(res.ref_end1 + 1):]

    return (bp_pos_reference, tconsensus_after)


def locate_bp_sbnd(consensus_file, output_file, reference_fasta, debug):

    fasta_file_ins = pysam.FastaFile(reference_fasta)

    with open(consensus_file, 'r') as hin, open(output_file, 'w') as hout, \
        open(output_file + ".locate_bp.sbnd.log", 'w') as hout_log:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            temp_key, tconsensus = F[0], F[1]
            print(temp_key + '\n' + tconsensus, file = hout_log)
            tchr, tstart, tend, tdir, _, tcid = temp_key.split(',')
            tstart, tend = int(tstart), int(tend)  
            bret = get_refined_bp_sbnd(tconsensus, fasta_file_ins, tchr, tstart, tend, tdir, hout_log)
            if bret is not None:  
                bp_pos, tconsensus_after = bret
                print(bp_pos, tconsensus_after, file = hout_log)
                print('', file = hout_log)
                print(f"{tchr}\t{bp_pos}\t{tdir}\t{tconsensus_after}\tb_{tcid}", file = hout)
                # print('\t'.join([tchr, str(bp_pos), tdir, tconsensus_after]), file = hout)

    if not debug: os.remove(output_file + ".locate_bp.sbnd.log")





#! /usr/bin/env python3

import csv, itertools
import pysam, parasail

from nanomonsv.my_seq import reverse_complement
from nanomonsv.logger import get_logger

logger = get_logger(__name__)

class Sv(object):

    def __init__(self, tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, 
        total_read_tumor = None, var_read_tumor = None, total_read_ctrl = None, var_read_ctrl = None, 
        insert_type = None, print_line = None):

        self.chr1 = tchr1
        self.pos1 = tpos1
        self.dir1 = tdir1
        self.chr2 = tchr2
        self.pos2 = tpos2
        self.dir2 = tdir2
        self.inseq = tinseq
        self.total_read_tumor = total_read_tumor
        self.var_read_tumor = var_read_tumor
        self.total_read_ctrl = total_read_ctrl
        self.var_read_ctrl = var_read_ctrl
        self.insert_type = insert_type
        self.print_line = print_line
        self.filter = []


class Duplicate_remover(object):

    def __init__(self, output_file, reference_fasta, simple_repeat_bed = None, 
        bp_dist_margin = 30, validate_seg_len = 100, simple_repeat_dist_margin = 50):
        self.sv_list = []
        self.hout = open(output_file, 'w')
        self.reference_h = pysam.FastaFile(reference_fasta)
        self.simple_repeat_tb = pysam.TabixFile(simple_repeat_bed) if simple_repeat_bed is not None else None
        self.bp_dist_margin = bp_dist_margin
        self.validate_seg_len = validate_seg_len
        self.simple_repeat_dist_margin = simple_repeat_dist_margin
        self.header = None

    def __del__(self):
        self.hout.close()
        self.reference_h.close()
        if self.simple_repeat_tb is not None: self.simple_repeat_tb.close()

    def add_sv(self, tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq,
        total_read_tumor, var_read_tumor, total_read_ctrl, var_read_ctrl, insert_type, print_line):
                     
        sv = Sv(tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, 
            total_read_tumor, var_read_tumor, total_read_ctrl, var_read_ctrl, insert_type, print_line)
        self.sv_list.append(sv)


    def filter_close_both_breakpoints(self, sv1, sv2, filter_item = "Duplicate_with_close_SV"):

        # only apply when the first or the second sv is not insertion
        if len(sv1.inseq) >= 100 or len(sv2.inseq) >= 100: return 

        if sv1.chr1 == sv2.chr1 and sv1.chr2 == sv2.chr2 and sv1.dir1 == sv2.dir1 and sv1.dir2 == sv2.dir2 and \
            abs(sv1.pos1 - sv2.pos1) < self.bp_dist_margin and abs(sv1.pos2 - sv2.pos2) < self.bp_dist_margin:

            if sv1.var_read_tumor is not None and sv2.var_read_tumor is not None:
                if sv1.var_read_tumor > sv2.var_read_tumor: 
                    sv1.filter.append(filter_item); return
                if sv2.var_read_tumor < sv1.var_read_tumor: 
                    sv2.filter.append(filter_item); return

            if len(sv1.inseq) < len(sv2.inseq):
                sv1.filter.append(filter_item); return
            if len(sv2.inseq) < len(sv1.inseq): 
                sv2.filter.append(filter_item); return

            if sv1.pos1 < sv2.pos1: 
                sv1.filter.append(filter_item); return
            if sv2.pos1 < sv1.pos1: 
                sv2.filter.append(filter_item); return

            if sv1.pos2 < sv2.pos2: 
                sv1.filter.append(filter_item); return
            if sv2.pos2 < sv1.pos2: 
                sv2.filter.append(filter_item) ; return

            sv2.filter.append(filter_item); return



    def filter_sv_insertion_match(self, sv, ins, filter_item = "Duplicate_with_insertion"):
    
        # only apply when the first sv is not insertion and the second sv is insertion type
        if len(sv.inseq) >= 100 or len(ins.inseq) < 100: return 

        if not (sv.chr1 == ins.chr1 and abs(sv.pos1 - ins.pos1) <= self.bp_dist_margin) and \
            not (sv.chr2 == ins.chr2 and abs(sv.pos2 - ins.pos2) <= self.bp_dist_margin):
            return 

        ins_seg = self.reference_h.fetch(ins.chr1, max(ins.pos1 - self.validate_seg_len - self.bp_dist_margin - 1, 0), ins.pos1 - 1)
        ins_seg = ins_seg + ins.inseq
        ins_seg = ins_seg + self.reference_h.fetch(ins.chr1, ins.pos2 - 1, ins.pos2 + self.validate_seg_len + self.bp_dist_margin - 1)

        if sv.dir1 == '+':
            tseq = self.reference_h.fetch(sv.chr1, max(sv.pos1 - self.validate_seg_len - 1, 0), sv.pos1 - 1)
        else:
            tseq = self.reference_h.fetch(sv.chr1, sv.pos1 - 1, sv.pos1 + self.validate_seg_len - 1) 
            tseq = reverse_complement(tseq)

        if sv.dir1 == '+':
            sv_seg = tseq + sv.inseq
        else:
            sv_seg = tseq + reverse_complement(sv.inseq)

        if sv.dir2 == '-':
            tseq = self.reference_h.fetch(sv.chr2, sv.pos2 - 1, sv.pos2 + self.validate_seg_len - 1)
        else:
            tseq = self.reference_h.fetch(sv.chr2, max(sv.pos2 - self.validate_seg_len - 1, 0), sv.pos2 - 1)
            tseq = reverse_complement(tseq)
 
        sv_seg = sv_seg + tseq

        user_matrix = parasail.matrix_create("ACGT", 2, -2)
        res = parasail.ssw(sv_seg, ins_seg, 3, 1, user_matrix)
        res_r = parasail.ssw(reverse_complement(sv_seg), ins_seg, 3, 1, user_matrix)

        if res.score1 > res_r.score1:
            match_ratio = float(res.score1) / (2 * (res.ref_end1 - res.ref_begin1 + 1))
            if match_ratio > 0.75:
                if res.read_begin1 < 0.1 * len(sv_seg) and res.read_end1 > 0.9 * len(sv_seg): 
                    sv.filter.append(filter_item)
                    return 
        else:
            match_ratio = float(res_r.score1) / (2 * (res_r.ref_end1 - res_r.ref_begin1 + 1))
            if match_ratio > 0.75:
                if res_r.read_begin1 < 0.1 * len(sv_seg) and res_r.read_end1 > 0.9 * len(sv_seg): 
                    sv.filter.append(filter_item)
                    return

        return


    def filter_dup_insertion(self, ins1, ins2, filter_item = "Duplicate_with_close_insertion"):

        # only apply when the first and the second sv is insertion type
        if len(ins1.inseq) < 100 or len(ins2.inseq) < 100: return None

        bp_match = (ins1.chr1 == ins2.chr1 and abs(ins1.pos1 - ins2.pos1) <= 2 * self.bp_dist_margin) and \
            (ins1.chr2 == ins2.chr2 and abs(ins1.pos2 - ins2.pos2) <= 2 * self.bp_dist_margin)

        if bp_match == False: return None

        # insertion with shorter length will be filtered
        if len(ins1.inseq) < len(ins2.inseq):
            ins1.filter.append(filter_item)
            return
        else:
            ins2.filter.append(filter_item)


    def filter_sv_with_decoy(self, sv, filter_item = "SV_with_decoy"):

        if sv.chr1.endswith("decoy") or sv.chr2.endswith("decoy"): 
            sv.filter.append(filter_item)
            return


    def filter_small_sv(self, sv, size_thres = 100, filter_item = "Too_small_SV"):

        if sv.chr1 == sv.chr2 and sv.dir1 == '+' and sv.dir2 == '-':
            sv_size = sv.pos2 - sv.pos1 + len(sv.inseq) - 1
            if sv_size < size_thres:
                sv.filter.append(filter_item)
                return

    def filter_indel_in_simple_repeat(self, sv, rescure_annotated_insertions = True, filter_item = "Simple_repeat"):

        if sv.chr1 == sv.chr2 and sv.dir1 == '+' and sv.dir2 == '-':
            sv_size = sv.pos2 - sv.pos1 + len(sv.inseq) - 1
 
            if sv.insert_type is not None and rescure_annotated_insertions: return

            tabix_error_flag = False
            try:
                records = self.simple_repeat_tb.fetch(sv.chr1, max(sv.pos1 - self.simple_repeat_dist_margin + 1, 0), 
                    sv.pos1 + self.simple_repeat_dist_margin)
            except Exception as inst:
                logger.warning("%s: %s" % (type(inst), inst.args))
                tabix_error_flag = True

            if tabix_error_flag == False:
                for record_line in records:
                    record = record_line.split('\t')
                    if sv.pos1 >= int(record[1]) - self.simple_repeat_dist_margin and \
                        int(sv.pos2) <= int(record[2]) + self.simple_repeat_dist_margin:
                        sv.filter.append(filter_item); return


    def apply_filters(self):

        # logger.info("filter_close_both_breakpoints")
        for sv1, sv2 in itertools.combinations(self.sv_list, 2):
            self.filter_close_both_breakpoints(sv1, sv2)

        # logger.info("filter_sv_insertion_match")
        for sv1, sv2 in itertools.combinations(self.sv_list, 2):

            if len(sv1.inseq) < 100 and len(sv2.inseq) >= 100: 
                self.filter_sv_insertion_match(sv1, sv2)
            elif len(sv2.inseq) < 100 and len(sv1.inseq) >= 100:
                self.filter_sv_insertion_match(sv2, sv1)

        # logger.info("filter_dup_insertion")
        for ins1, ins2 in itertools.combinations(self.sv_list, 2):

            if len(ins1.inseq) >= 100 and len(ins2.inseq) >= 100:
                self.filter_dup_insertion(ins1, ins2)

        for sv in self.sv_list:
            self.filter_sv_with_decoy(sv)

        for sv in self.sv_list:
            self.filter_small_sv(sv)

        if self.simple_repeat_tb is not None:
            for sv in self.sv_list:
                self.filter_indel_in_simple_repeat(sv)
    

    def flush_sv_list(self):

        print(self.header + "\tIs_filter", file = self.hout)
        for sv in self.sv_list:
            if len(sv.filter) == 0: 
                filter_print = "PASS"
            else:
                filter_print = ';'.join(list(set(sv.filter)))

            print(f"{sv.print_line}\t{filter_print}", file = self.hout)


def post_filter_main(args):

    duplicate_remover = Duplicate_remover(args.output_file, args.reference_fasta, simple_repeat_bed = args.simple_repeat_bed)

    with open(args.sv_list_file, 'r') as hin:
        dreader = csv.DictReader(hin, delimiter = '\t')
        header = dreader.fieldnames
        duplicate_remover.header = '\t'.join(header)
        if "Checked_Read_Num_Control" in header and "Supporting_Read_Num_Control" in header: is_control = True
        if "Insert_Type" in header: is_insert_type = True

        for F in dreader:
            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = F["Chr_1"], int(F["Pos_1"]), F["Dir_1"], F["Chr_2"], int(F["Pos_2"]), F["Dir_2"], F["Inserted_Seq"]

            total_read_tumor, var_read_tumor = int(F["Checked_Read_Num_Tumor"]), int(F["Supporting_Read_Num_Tumor"])
            if is_control:
                total_read_ctrl, var_read_ctrl = int(F["Checked_Read_Num_Control"]), int(F["Supporting_Read_Num_Control"])
            else:
                total_read_ctrl, var_read_ctrl = None, None

            if is_insert_type:
                insert_type = F["Insert_Type"] if F["Insert_Type"] not in ["None", "---"] else None
            else:
                insert_type = None

            """
            # filtering by supporting read and variant frequencies
            if total_read_tumor == 0: continue
            if var_read_tumor < min_tumor_variant_read_num: continue
            if float(var_read_tumor) / float(total_read_tumor) < min_tumor_VAF: continue

            if is_control:
                if total_read_ctrl is None or total_read_ctrl == 0: continue
                if var_read_ctrl > max_control_variant_read_num: continue
                if float(var_read_ctrl) / float(total_read_ctrl) > max_control_VAF: continue
            """

            duplicate_remover.add_sv(tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq,
                total_read_tumor = total_read_tumor, var_read_tumor = var_read_tumor, 
                total_read_ctrl = total_read_ctrl, var_read_ctrl = var_read_ctrl, 
                insert_type = insert_type, print_line = '\t'.join(F.values()))


    duplicate_remover.apply_filters()
    duplicate_remover.flush_sv_list()
    del duplicate_remover


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(prog = "nanomonsv_post_filter",
        description = "Post filtering script for the result of nanomonsv")

    parser.add_argument("sv_list_file", type = str,
                        help = "Path to the nanomonsv result file")

    parser.add_argument("output_file", type = str,
                        help = "Path to the output file")

    parser.add_argument("reference_fasta", metavar = "reference.fa", type = str,
                        help = "Path to the reference genome sequence")

    parser.add_argument("--simple_repeat_bed", metavar = "simpleRepeat.bed.gz", type = str, default = None,
                        help = "Path to the tabix indexed simple repeat bed file")

    args = parser.parse_args()

    post_filter_main(args)





#i! /usr/bin/env python

import sys, gzip, statistics
import pysam
from .logger import get_logger

def cluster_rearrangement(input_file, output_file, cluster_margin_size = 100):

    hout = open(output_file, 'w')

    merged_bedpe = {}
    with gzip.open(input_file, 'rt') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
    
            for key in sorted(merged_bedpe):

                bchr1, bstart1, bend1, bdir1, bchr2, bstart2, bend2, bdir2 = key
                breadids, binfo1, binfo2 = merged_bedpe[key]

                if F[0] != bchr1 or int(F[1]) > int(bend1) + cluster_margin_size: 

                    print('\t'.join([bchr1, str(bstart1), str(bend1), bchr2, str(bstart2), str(bend2), \
                       breadids, '0', bdir1, bdir2, binfo1, binfo2]), file = hout)
                    del merged_bedpe[key]


            match_flag = False
            for key in sorted(merged_bedpe):

                bchr1, bstart1, bend1, bdir1, bchr2, bstart2, bend2, bdir2 = key 
                breadids, binfo1, binfo2 = merged_bedpe[key]

                if F[6] in breadids: continue

                if F[0] == bchr1 and F[3] == bchr2 and F[8] == bdir1 and F[9] == bdir2 and \
                  int(F[2]) >= bstart1 and int(F[1]) <= bend1 and int(F[5]) >= bstart2 and int(F[4]) <= bend2:
                        
                    new_start1 = min(int(F[1]), bstart1)
                    new_end1 = max(int(F[2]), bend1)
                    new_start2 = min(int(F[4]), bstart2)
                    new_end2 = max(int(F[5]), bend2)
                            
                    new_key = (bchr1, new_start1, new_end1, bdir1, bchr2, new_start2, new_end2, bdir2)
                    new_readids = breadids + ';' + F[6]
                    new_info1 = binfo1 + ';' + F[10]
                    new_info2 = binfo2 + ';' + F[11]

                    del merged_bedpe[key]

                    merged_bedpe[new_key] = (new_readids, new_info1, new_info2)
                    match_flag = True
                    break

                    
            if not match_flag:
                new_key = (F[0], int(F[1]), int(F[2]), F[8], F[3], int(F[4]), int(F[5]), F[9])
                merged_bedpe[new_key] = (F[6], F[10], F[11])


    for key in sorted(merged_bedpe):
        bchr1, bstart1, bend1, bdir1, bchr2, bstart2, bend2, bdir2 = key
        breadids, binfo1, binfo2 = merged_bedpe[key]

        print('\t'.join([bchr1, str(bstart1), str(bend1), bchr2, str(bstart2), str(bend2), \
          breadids, '0', bdir1, bdir2, binfo1, binfo2]), file = hout)

    hout.close()


def filt_clustered_rearrangement1(input_file, output_file, read_num_thres = 3, median_mapQ_thres = 40, max_overhang_size_thres = 300):

    hout = open(output_file, 'w') 
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            read_ids = F[6].split(';')
            if len(read_ids) < read_num_thres: continue

            info1 = F[10].split(';')
            info2 = F[11].split(';')

            median_mapQ1 = statistics.median([int(x.split(',')[5]) for x in info1])
            median_mapQ2 = statistics.median([int(x.split(',')[5]) for x in info2])
        
            max_overhang_size1 = max([abs(int(x.split(',')[2]) - int(x.split(',')[0])) for x in info1])
            max_overhang_size2 = max([abs(int(x.split(',')[2]) - int(x.split(',')[0])) for x in info2]) 
        
            if median_mapQ1 < median_mapQ_thres or median_mapQ2 < median_mapQ_thres: continue
            if max_overhang_size1 < max_overhang_size_thres or max_overhang_size2 < max_overhang_size_thres: continue

            print('\t'.join(F), file = hout)


    hout.close()



def filt_clustered_rearrangement2(input_file, output_file, control_junction_bedpe, control_read_num_thres = 0, control_check_margin = 50):

    hout = open(output_file, 'w')
    if control_junction_bedpe is not None: control_junction_db = pysam.TabixFile(control_junction_bedpe)
    
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tchr1, tstart1, tend1, tchr2, tstart2, tend2 = F[0], F[1], F[2], F[3], F[4], F[5]
            tdir1, tdir2 = F[8], F[9]

            is_short_deletion = True if tchr1 == tchr2 and tdir1 == '+' and tdir2 == '-' and int(tend2) - int(tstart1) <= 500 else False

            control_flag = False
            if control_junction_bedpe is not None:
                tabix_error_flag = False
                try:
                    records = control_junction_db.fetch(F[0], max(0, int(tstart1) - 200), int(tend1) + 200)
                except:
                    tabix_error_flag = True

                if not tabix_error_flag:
                    for record_line in records:
                        record = record_line.split('\t')

                        if is_short_deletion:
                            if tchr1 == record[0] and tchr2 == record[3] and tdir1 == record[8] and tdir2 == record[9] and \
                                int(tstart1) <= int(record[5]) and int(tend2) >= int(record[1]):   
                                control_flag = True
                        else:
                            if tchr1 == record[0] and tdir1 == record[8] and int(tend1) >= int(record[1]) - control_check_margin and int(tstart1) <= int(record[2]) + control_check_margin and \
                                tchr2 == record[3] and tdir2 == record[9] and int(tend2) >= int(record[4]) - control_check_margin and int(tstart2) <= int(record[5]) + control_check_margin:
                                control_flag = True

            if not control_flag:
                print('\t'.join(F), file = hout)


    hout.close()
    if control_junction_bedpe is not None: control_junction_db.close()



def cluster_insertion_deletion(input_file, output_file, deletion_cluster_margin_size = 10, check_margin_size = 50, size_margin_ratio = 0.2, maximum_unique_pairs = 100, maximum_local_variant_num = 1000):

    logger = get_logger(__name__)

    dcms = deletion_cluster_margin_size
    hout = open(output_file, 'w')
    
    merged_bedpe = {}
    skip_pos = 0
    tmp_pos = 0
    tmp_chr = None
    local_variant_num = 0

    skipped_pos_list = []

    with gzip.open(input_file, 'rt') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            if F[0] == "hs37d5": continue

            if F[0] != tmp_chr: tmp_chr, tmp_pos, skip_pos = F[0], 0,  0
            if int(F[1]) < skip_pos: continue
            if int(F[4]) < 90: continue
 
            for key in sorted(merged_bedpe):

                # bchr, bstart, bend = key
                bchr1, bstart1, bend1, bdir1, bchr2, bstart2, bend2, bdir2 = key
                breadids, bsize, binfo = merged_bedpe[key]

                if F[0] != bchr1 or int(F[1]) > int(bend1) + check_margin_size: 

                    print('\t'.join([bchr1, str(bstart1), str(bend1), bchr2, str(bstart2), str(bend2), \
                       breadids, '0', bdir1, bdir2, bsize, binfo]), file = hout)
                    del merged_bedpe[key]

                    tmp_pos = int(F[1])
                    local_variant_num = 0

            match_flag = False
            for key in sorted(merged_bedpe):

                # bchr, bstart, bend = key 
                bchr1, bstart1, bend1, bdir1, bchr2, bstart2, bend2, bdir2 = key 
                breadids, bsize, binfo = merged_bedpe[key]
                
                bsize_vec = [float(x) for x in bsize.split(';')]

                # if F[0] == bchr and abs(int(F[1]) - int(bstart)) <= deletion_cluster_margin_size and \
                #     abs(int(F[2]) - int(bend)) <= deletion_cluster_margin_size:
                # if F[0] == bchr1 and F[3] == bchr2 and F[8] == bdir1 and F[9] == bdir2 and \
                #   int(F[2]) >= bstart1 and int(F[1]) <= bend1 and int(F[5]) >= bstart2 and int(F[4]) <= bend2:
                if F[0] == bchr1 and int(F[1]) - dcms <= bend1 and int(F[1]) + dcms >= bstart1 and \
                  int(F[2]) - dcms <= bend2 and int(F[2]) + dcms >= bstart2 and \
                  float(F[4]) > (1 - size_margin_ratio) * min(bsize_vec) and float(F[4]) < (1 + size_margin_ratio) * max(bsize_vec):

                    new_start1 = min(int(F[1]) - dcms, bstart1)
                    new_end1 = max(int(F[1]) + dcms, bend1)
                    new_start2 = min(int(F[2]) - dcms, bstart2)
                    new_end2 = max(int(F[2]) + dcms, bend2)
                            
                    new_key = (bchr1, new_start1, new_end1, bdir1, bchr2, new_start2, new_end2, bdir2)

                    # ney_key = key
                    new_size = bsize + ';' + F[4]
                    new_readids = breadids + ';' + F[3]
                    new_info = binfo + ';' + F[6]

                    del merged_bedpe[key]

                    merged_bedpe[new_key] = (new_readids, new_size, new_info)
                    match_flag = True
                    break

                    
            if not match_flag:
                new_key = (F[0], int(F[1]) - dcms, int(F[1]) + dcms, '+', F[0], int(F[2]) - dcms, int(F[2]) + dcms, '-') 
                # new_key = (F[0], F[1], F[2])
                merged_bedpe[new_key] = (F[3], F[4], F[6])

            local_variant_num = local_variant_num + 1

            if local_variant_num >= maximum_local_variant_num:
                skipped_pos_list.append(F[0] + ':' + F[1])
                merged_bedpe = {}
                local_variant_num = 0
                skip_pos = int(F[1]) + 100 * check_margin_size

            if len(merged_bedpe) > maximum_unique_pairs:
                skipped_pos_list.append(F[0] + ':' + F[1])
                merged_bedpe = {}
                local_variant_num = 0
                skip_pos = int(F[1]) + 100 * check_margin_size

            
            

        for key in sorted(merged_bedpe):
            # bchr, bstart, bene = key
            bchr1, bstart1, bend1, bdir1, bchr2, bstart2, bend2, bdir2 = key
            breadids, bsize, binfo = merged_bedpe[key]

            print('\t'.join([bchr1, str(bstart1), str(bend1), bchr2, str(bstart2), str(bend2), \
              breadids, '0', bdir1, bdir2, bsize, binfo]), file = hout)

            # if F[0] != bchr or int(F[1]) > int(bend) + check_margin_size:  
            #     print('\t'.join([bchr, str(bstart), str(bend), breadids, bsize, '+', binfo]), file = hout)

    skipped_pls_list_line = ' '.join(skipped_pos_list)
    logger.info(f"Because of too many local variants, clustering procedures were skipped around: {skipped_pls_list_line}")

    hout.close()



def filt_clustered_insertion_deletion1(input_file, output_file, bp_bed, read_num_thres = 3, median_mapQ_thres = 40, max_overhang_size_thres = 300):

    hout = open(output_file, 'w')
    if bp_bed is not None: bp_bed_db = pysam.TabixFile(bp_bed)

    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tchr, tstart1, tend1, tstart2, tend2 = F[0], int(F[1]), int(F[2]), int(F[4]), int(F[5])
            read_ids = list(set(F[6].split(';')))

            # add breakpoint info
            if bp_bed is not None:
                tabix_error_flag = False
                try:
                    records = bp_bed_db.fetch(tchr, max(0, tstart1), tend2)
                except:
                    tabix_error_flag = True

                if not tabix_error_flag:
                    for record_line in records:
                        record = record_line.split('\t')
                        if (tchr == record[0] and tstart1 <= int(record[2]) and int(record[2]) <= tend1 and record[5] == '+') or \
                            (tchr == record[0] and tstart2 <= int(record[2]) and int(record[2]) <= tend2 and record[5] == '-'):
   
                            if record[3] in F[6]: continue
                            F[6] = F[6] + ';' + record[3]
                            F[10] = F[10] + ';' + record[5]
                            F[11] = F[11] + ';' + record[6]

            read_ids = list(set(F[6].split(';')))    
            if len(read_ids) < read_num_thres: continue

            # F[6] = ';'.join(read_ids)
            info = F[11].split(';')

            median_mapQ = statistics.median([int(x.split(',')[5]) for x in info])
            non_secondary_readnum = len([x.split(',')[10] for x in info if x.split(',')[10] == "False"])
            # max_overhang_size = max([abs(int(x.split(',')[1]) - int(x.split(',')[0])) for x in info2])

            if median_mapQ < median_mapQ_thres: continue
            if non_secondary_readnum < read_num_thres: continue
            # if max_overhang_size1 < max_overhang_size_thres or max_overhang_size2 < max_overhang_size_thres: continue

            print('\t'.join(F), file = hout)


    hout.close()


def filt_clustered_insertion_deletion2(input_file, output_file, control_junction_bedpe, control_read_num_thres = 0, control_check_margin = 50):

    hout = open(output_file, 'w')
    if control_junction_bedpe is not None: control_junction_db = pysam.TabixFile(control_junction_bedpe)
    
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            tchr1, tstart1, tend1, tchr2, tstart2, tend2 = F[0], F[1], F[2], F[3], F[4], F[5]
            median_size = statistics.median([int(x) for x in F[10].split(';') if x not in ['-', '+']])

            control_flag = False
            if control_junction_bedpe is not None:
                tabix_error_flag = False
                try:
                    records = control_junction_db.fetch(F[0], max(0, int(tstart1) - 50), int(tend2) + 50)
                except:
                    tabix_error_flag = True

                if not tabix_error_flag:
                    for record_line in records:
                        record = record_line.split('\t')

                        if tchr1 == record[0] and int(tstart1) - control_check_margin <= int(record[2]) and int(tend2) + control_check_margin >= int(record[1]) \
                            and int(record[4]) >= median_size * 0.5:
                            # and int(record[2]) - int(record[1]) >= median_size * 0.5:
                            control_flag = True


            if not control_flag:
                print('\t'.join(F), file = hout)


    hout.close()
    if control_junction_bedpe is not None: control_junction_db.close()


def filt_final(input_file, input_sread_file, output_file, output_sread_file, min_tumor_variant_read_num = 3, min_tumor_VAF = 0.05, max_control_variant_read_num = 1, max_control_VAF = 0.03, is_no_filt = False, is_control = False):

    hout = open(output_file, 'w')
    header = ["Chr_1", "Pos_1", "Dir_1", "Chr_2", "Pos_2", "Dir_2", "Inserted_Seq", "Checked_Read_Num_Tumor", "Supporting_Read_Num_Tumor"]
    if is_control: header = header + ["Checked_Read_Num_Control", "Supporting_Read_Num_Control"]

    key2ikey = {}    
    print('\t'.join(header), file = hout)
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            checked_read_num_tumor, supporting_read_num_tumor = int(F[7]), int(F[8])
            if is_control: checked_read_num_control, supporting_read_num_control = int(F[9]), int(F[10])

            if not is_no_filt:
                if checked_read_num_tumor == 0: continue
                if supporting_read_num_tumor < min_tumor_variant_read_num: continue
                if float(supporting_read_num_tumor) / float(checked_read_num_tumor) < min_tumor_VAF: continue

                if is_control:
                    if checked_read_num_control == 0: continue
                    if supporting_read_num_control > max_control_variant_read_num: continue
                    if float(supporting_read_num_control) / float(checked_read_num_control) > max_control_VAF: continue

            print('\t'.join(F), file = hout)

            key = ','.join(F[:6]) + ',' +  str(len(F[6] if F[6] != "---" else ''))
            ikey = ','.join(F[:7])
            key2ikey[key] = ikey

    hout.close()

    
    hout = open(output_sread_file, 'w') 
    with open(input_sread_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key = F[0]
            read_id = F[1]

            score1, cstart1, cend1, sstart1, send1, strand1 = int(F[2]), int(F[3]), int(F[4]), int(F[5]), int(F[6]), F[7]
            score2, cstart2, cend2, sstart2, send2, strand2 = int(F[8]), int(F[9]), int(F[10]), int(F[11]), int(F[12]), F[13]

            if key in key2ikey:

                ikey = key2ikey[key]
                tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = ikey.split(',')
                
                if score1 >= score2:
                    sstrand = strand1
                    if sstrand == '+':
                        spos = (float(send1 - sstart1) / float(cend1 - cstart1)) * (200 - cstart1) + sstart1
                    else:
                        spos = (float(sstart1 - send1) / float(cend1 - cstart1)) * (200 - cstart1) + send1
                else:
                    sstrand = strand2
                    if sstrand == '+':
                        spos = (float(send1 - sstart1) / float(cend1 - cstart1)) * (200 - cstart1) + sstart1 - len(tinseq) - 1
                    else:
                        spos = (float(sstart1 - send1) / float(cend1 - cstart1)) * (200 - cstart1) + send1 + len(tinseq) + 1
                  
                print(key2ikey[key] + '\t' + read_id + '\t' + str(int(spos)) + '\t' + sstrand, file = hout)


    hout.close()



"""
if __name__ == "__main__":

    import sys

    tumor_prefix = sys.argv[1]
    control_prefix = sys.argv[2]

    cluster_junction(tumor_prefix + ".junction.sorted.bedpe.gz", tumor_prefix + ".junction.sorted.clustered.bedpe")

    filt_clustered_junction1(tumor_prefix + ".junction.sorted.clustered.bedpe", tumor_prefix + ".junction.sorted.clustered.filt1.bedpe")

    filt_clustered_junction2(tumor_prefix + ".junction.sorted.clustered.filt1.bedpe", tumor_prefix + ".junction.sorted.clustered.filt2.bedpe", 
                             control_prefix + ".junction.sorted.bedpe.gz")
"""




#! /usr/bin/en python3

import sys, csv, datetime
import pysam
from .version import __version__
from .my_seq import reverse_complement


def genomesv2vcf_convert(result_file, output_vcf, reference):

    today_str = datetime.datetime.today().strftime("%Y%m%d")

    header = '##fileformat=VCFv4.3\n'\
             f'##fileDate={today_str}\n'\
             f'##source=nanomonsv-{__version__}\n'\
             f'##reference={reference}'

    ref_tb = pysam.FastaFile(reference)

    for (tchr, tlen) in zip(ref_tb.references, ref_tb.lengths):
        header = header + '\n' + f"##contig=<ID={tchr},length={tlen}>"

    header = header + '\n' + \
            '##FILTER=<ID=Duplicate_with_close_SV,Description="When multiple SVs that share breakpoints in close proximity are detected, all but one SVs are filtered.">\n'\
            '##FILTER=<ID=Duplicate_with_insertion,Description="Breakend SVs that are inferred to be the same as any of detected insertions">\n'\
            '##FILTER=<ID=Duplicate_with_close_insertion,Description="When multiple insertions in close proximity are detected, all but one insertions are filtered.">\n'\
            '##FILTER=<ID=SV_with_decoy,Description="SVs involving decoy contigs">\n'\
            '##FILTER=<ID=Too_small_size,Description="Insertions whose size is below the threshould (currently 100bp)">\n'\
            '##FILTER=<ID=Too_low_VAF,Description="SVs whose variant allele frequencies are inferred to be low">\n'\
            '##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">\n'\
            '##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Difference in length between REF and ALT alleles">\n'\
            '##INFO=<ID=END,Number=1,Type=Integer,Description="End position of the variant described in this record">\n'\
            '##INFO=<ID=MATEID,Number=1,Type=String,Description="ID of mate breakend">\n'\
            '##INFO=<ID=SVINSLEN,Number=1,Type=Integer,Description="Length of insertion">\n'\
            '##INFO=<ID=SVINSSEQ,Number=1,Type=String,Description="Sequence of insertion">\n'\
            '##ALT=<ID=DEL,Description="Deletion">\n'\
            '##ALT=<ID=INS,Description="Insertion">\n'\
            '##ALT=<ID=DUP,Description="Duplication">\n'\
            '##FORMAT=<ID=TR,Number=1,Type=Integer,Description="The number of reads around the breakpoints">\n'\
            '##FORMAT=<ID=VR,Number=1,Type=Integer,Description="The number of variant supporting reads determined in the validation realignment step">'

    chrom_order = {chrom: i for i, chrom in enumerate(ref_tb.references)}

    with open(result_file, 'r') as hin, open(output_vcf, 'w') as hout:

        dreader = csv.DictReader(hin, delimiter = '\t')
        fieldname_list = dreader.fieldnames
        is_control = True if "Checked_Read_Num_Control" in fieldname_list and "Supporting_Read_Num_Control" in fieldname_list else False

        if is_control:
            header = header + '\n' + "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tTUMOR\tCONTROL"
        else:
            header = header + '\n' + "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tTUMOR"

        records = []
        for F in dreader:

            tchrom = F["Chr_1"]
            tid = F["SV_ID"]
            tqual = '.'
            tfilter = F["Is_Filter"]

            if F["Inserted_Seq"] != "---":
                tsvinsseq = F["Inserted_Seq"]
                tsvinslen = len(F["Inserted_Seq"])
            else:
                tsvinsseq = ''
                tsvinslen = 0
            
            tformat_sample = f'TR:VR\t{F["Checked_Read_Num_Tumor"]}:{F["Supporting_Read_Num_Tumor"]}'
            if is_control:
                tformat_sample = tformat_sample + f'\t{F["Checked_Read_Num_Control"]}:{F["Supporting_Read_Num_Control"]}' 

            if F["Chr_1"] == F["Chr_2"] and F["Dir_1"] == '+' and F["Dir_2"] == '-':

                tpos = int(F["Pos_1"])
                tref = ref_tb.fetch(tchrom, tpos - 1, tpos)
                if tref == '' or tref is None: continue
                tsvlen = int(F["Pos_2"]) - int(F["Pos_1"]) - 1
                tend = int(F["Pos_2"]) - 1
    
                # Deletion
                if tsvlen > tsvinslen:
                    talt = "<DEL>"
                    tsvlen = int(F["Pos_2"]) - int(F["Pos_1"]) - 1
                    tinfo = f"END={tend};SVTYPE=DEL;SVLEN=-{tsvlen}"

                    if tsvinslen != 0:
                        tinfo = tinfo + f";SVINSLEN={tsvinslen};SVINSSEQ={tsvinsseq}"

                # Insertion
                else:
                    talt = "<INS>"
                    tinfo = f"END={tend};SVTYPE=INS;SVINSLEN={tsvinslen};SVINSSEQ={tsvinsseq}"

                records.append((chrom_order.get(tchrom, float('inf')), tpos, f"{tchrom}\t{tpos}\t{tid}\t{tref}\t{talt}\t{tqual}\t{tfilter}\t{tinfo}\t{tformat_sample}"))

            # Duplication
            elif F["Chr_1"] == F["Chr_2"] and F["Dir_1"] == '-' and F["Dir_2"] == '+' and F["Pos_1"] != '1': 

                tpos = int(F["Pos_1"]) - 1
                tref = ref_tb.fetch(tchrom, tpos - 1, tpos)
                if tref == '' or tref is None: continue
                talt = "<DUP>"
                tend = int(F["Pos_2"]) 
                tsvlen = int(F["Pos_2"]) - int(F["Pos_1"]) + 1
                tinfo = f"END={tend};SVTYPE=DUP;SVLEN={tsvlen}"
                if tsvinslen != 0:
                    tinfo = tinfo + f";SVINSLEN={tsvinslen};SVINSSEQ={tsvinsseq}"

                records.append((chrom_order.get(tchrom, float('inf')), tpos, f"{tchrom}\t{tpos}\t{tid}\t{tref}\t{talt}\t{tqual}\t{tfilter}\t{tinfo}\t{tformat_sample}"))

            # Breakend
            #
            # VCF BND ALT field encoding
            #
            # Example: chr1:100 (REF=A) ↔ chr3:500 (REF=G)
            #
            # Record 1 (chr1:100 side):
            #   Dir_1  Dir_2  ALT             Note
            #   +      -      A[chr3:500[     REF left,  bracket [ (Dir_2=-)
            #   +      +      A]chr3:500]     REF left,  bracket ] (Dir_2=+)
            #   -      -      [chr3:500[A     REF right, bracket [ (Dir_2=-)
            #   -      +      ]chr3:500]A     REF right, bracket ] (Dir_2=+)
            #
            # Record 2 (chr3:500 side, mate):
            #   Dir_1  Dir_2  ALT             Note
            #   +      -      ]chr1:100]G     REF right, bracket ] (Dir_1=+)
            #   +      +      G]chr1:100]     REF left,  bracket ] (Dir_1=+)
            #   -      -      [chr1:100[G     REF right, bracket [ (Dir_1=-)
            #   -      +      G[chr1:100[     REF left,  bracket [ (Dir_1=-)
            #
            # Rules:
            #   Record 1: Dir_1=+ -> REF left,  Dir_1=- -> REF right
            #             Dir_2=+ -> ']',       Dir_2=- -> '['
            #   Record 2: Dir_2=+ -> REF left,  Dir_2=- -> REF right
            #             Dir_1=+ -> ']',       Dir_1=- -> '['
            #
            # With insertion (SVINSSEQ, e.g. "TCG"):
            #   Inserted sequence is placed between the REF base and the bracketed mate.
            #   Example:
            #     Record 1, Dir_1=+, Dir_2=- : ATCG[chr3:500[
            #     Record 1, Dir_1=-, Dir_2=+ : ]chr3:500]TCGA
            #
            #   For the mate record, if the same inserted junction sequence is represented
            #   from the opposite breakend, use the reverse-complemented inserted sequence.
            #
            else:
                tchrom1 = F["Chr_1"]
                tpos1 = int(F["Pos_1"])
                tref1 = ref_tb.fetch(tchrom1, tpos1 - 1, tpos1)
                if tref1 == '' or tref1 is None: continue

                tchrom2 = F["Chr_2"]
                tpos2 = int(F["Pos_2"])
                tref2 = ref_tb.fetch(tchrom2, tpos2 - 1, tpos2)
                if tref2 == '' or tref2 is None: continue

                tbracket = ']' if F["Dir_2"] == '+' else '['
                if F["Dir_1"] == '+':
                    talt1 = f'{tref1}{tsvinsseq}{tbracket}{tchrom2}:{tpos2}{tbracket}'
                else:
                    talt1 = f'{tbracket}{tchrom2}:{tpos2}{tbracket}{tsvinsseq}{tref1}'

                tinfo1 = f"SVTYPE=BND;MATEID={tid}_1"
                if tsvinslen != 0: tinfo1 = tinfo1 + f";SVINSLEN={tsvinslen};SVINSSEQ={tsvinsseq}"

                records.append((chrom_order.get(tchrom1, float('inf')), tpos1, f"{tchrom1}\t{tpos1}\t{tid}_0\t{tref1}\t{talt1}\t{tqual}\t{tfilter}\t{tinfo1}\t{tformat_sample}"))

                # tchrom2 = F["Chr_2"]
                # tpos = int(F["Pos_2"])
                # tref = ref_tb.fetch(tchrom2, tpos - 1, tpos)
                # if tref == '' or tref is None: continue
                tbracket = ']' if F["Dir_1"] == '+' else '['
                tsvinsseq = reverse_complement(tsvinsseq)
                if F["Dir_2"] == '+':
                    talt2 = f'{tref2}{tsvinsseq}{tbracket}{tchrom1}:{tpos1}{tbracket}'
                else:
                    talt2 = f'{tbracket}{tchrom1}:{tpos1}{tbracket}{tsvinsseq}{tref2}'

                tinfo2 = f"SVTYPE=BND;MATEID={tid}_0"
                if tsvinslen != 0: tinfo2 = tinfo2 + f";SVINSLEN={tsvinslen};SVINSSEQ={tsvinsseq}"

                records.append((chrom_order.get(tchrom2, float('inf')), tpos2, f"{tchrom2}\t{tpos2}\t{tid}_1\t{tref2}\t{talt2}\t{tqual}\t{tfilter}\t{tinfo2}\t{tformat_sample}"))

        # Add Simple_repeat FILTER header if any record uses it
        if any('Simple_repeat' in line for _, _, line in records):
            header = header.replace(
                '##INFO=<ID=SVTYPE',
                '##FILTER=<ID=Simple_repeat,Description="SVs overlapping with simple repeat regions">\n##INFO=<ID=SVTYPE'
            )

        print(header, file = hout)

        records.sort()
        for _, _, line in records:
            print(line, file = hout)


def sbnd2vcf_convert(sbnd_result_file, output_vcf, reference):

    today_str = datetime.datetime.today().strftime("%Y%m%d")

    header = '##fileformat=VCFv4.3\n'\
             f'##fileDate={today_str}\n'\
             f'##source=nanomonsv-{__version__}\n'\
             f'##reference={reference}'

    ref_tb = pysam.FastaFile(reference)

    for (tchr, tlen) in zip(ref_tb.references, ref_tb.lengths):
        header = header + '\n' + f"##contig=<ID={tchr},length={tlen}>"

    header = header + '\n' + \
            '##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">\n'\
            '##FORMAT=<ID=TR,Number=1,Type=Integer,Description="The number of reads around the breakpoints">\n'\
            '##FORMAT=<ID=VR,Number=1,Type=Integer,Description="The number of variant supporting reads determined in the validation realignment step">'

    chrom_order = {chrom: i for i, chrom in enumerate(ref_tb.references)}

    with open(sbnd_result_file, 'r') as hin, open(output_vcf, 'w') as hout:

        dreader = csv.DictReader(hin, delimiter = '\t')
        fieldname_list = dreader.fieldnames
        is_control = True if "Checked_Read_Num_Control" in fieldname_list and "Supporting_Read_Num_Control" in fieldname_list else False

        if is_control:
            header = header + '\n' + "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tTUMOR\tCONTROL"
        else:
            header = header + '\n' + "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tTUMOR"

        records = []
        for F in dreader:

            tchrom = F["Chr_1"]
            tpos = int(F["Pos_1"])
            tid = F["SV_ID"]
            tqual = '.'
            tfilter = F["Is_Filter"]
            tcontig = F["Contig"]

            tref = ref_tb.fetch(tchrom, tpos - 1, tpos)
            if tref == '' or tref is None: continue

            # Single breakend ALT field (VCF 4.3)
            #   Dir=+: {REF}{Contig}.   (breakpoint extends rightward)
            #   Dir=-: .{RC(Contig)}{REF} (sequence approaches from left, contig stored in RC orientation)
            if F["Dir_1"] == '+':
                talt = f'{tref}{tcontig}.'
            else:
                talt = f'.{reverse_complement(tcontig)}{tref}'

            tinfo = "SVTYPE=BND"

            tformat_sample = f'TR:VR\t{F["Checked_Read_Num_Tumor"]}:{F["Supporting_Read_Num_Tumor"]}'
            if is_control:
                tformat_sample = tformat_sample + f'\t{F["Checked_Read_Num_Control"]}:{F["Supporting_Read_Num_Control"]}'

            records.append((chrom_order.get(tchrom, float('inf')), tpos, f"{tchrom}\t{tpos}\t{tid}\t{tref}\t{talt}\t{tqual}\t{tfilter}\t{tinfo}\t{tformat_sample}"))

        # Add FILTER headers dynamically based on records
        filter_headers = []
        if any('Simple_repeat' in line for _, _, line in records):
            filter_headers.append('##FILTER=<ID=Simple_repeat,Description="Single breakends overlapping with simple repeat regions">')
        if any('Canonical_SV_overlap' in line for _, _, line in records):
            filter_headers.append('##FILTER=<ID=Canonical_SV_overlap,Description="Single breakends overlapping with canonical SV breakpoints">')
        if filter_headers:
            header = header.replace(
                '##INFO=<ID=SVTYPE',
                '\n'.join(filter_headers) + '\n##INFO=<ID=SVTYPE'
            )

        print(header, file = hout)

        records.sort()
        for _, _, line in records:
            print(line, file = hout)


if __name__ == "__main__":

    import sys
    result_file = sys.argv[1]
    output_vcf = sys.argv[2]
    reference = sys.argv[3]
    genomesv2vcf_convert(result_file, output_vcf, reference)



#! /usr/bin/env python3

import sys, gzip, os, subprocess
from .logger import get_logger

logger = get_logger(__name__)


def merge_bed_by_sv(bedpe_file, hout, prefix_label, is_bedpe = False):

    temp_key, temp_count = None, 0
    with gzip.open(bedpe_file, 'rt') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            if is_bedpe:
                key = '\t'.join(F[:6] + ['*'] + F[7:10])
            else:
                key = '\t'.join(F[:3] + ['*'] + F[4:6])
            if key != temp_key:
                if temp_key is not None:
                    print(f'{temp_key}\t{temp_count}\t{prefix_label}', file = hout)
                temp_key = key 
                temp_count = 0
            temp_count = temp_count + 1

        if temp_key is not None:
            print(f'{temp_key}\t{temp_count}\t{prefix_label}', file = hout)


def merge_bed_by_sample(input_bedpe, output_bedpe, is_bedpe = False):

    temp_key, temp_counts, temp_samples = None, [], []
    with open(input_bedpe, 'r') as hin, open(output_bedpe, 'w') as hout:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            if is_bedpe:
                key = '\t'.join(F[:10])
            else:
                key = '\t'.join(F[:6])
            if key != temp_key:
                if temp_key is not None:
                    temp_counts_print_line = ','.join(temp_counts)
                    temp_samples_print_line = ','.join(temp_samples)
                    print(f'{temp_key}\t{temp_counts_print_line}\t{temp_samples_print_line}', file = hout)
                temp_key, temp_counts, temp_samples = key, [], []

            if is_bedpe:
                temp_counts.append(F[10])
                temp_samples.append(F[11])
            else:
                temp_counts.append(F[6])
                temp_samples.append(F[7])

        if temp_key is not None:
            temp_counts_print_line = ','.join(temp_counts)
            temp_samples_print_line = ','.join(temp_samples)
            print(f'{temp_key}\t{temp_counts_print_line}\t{temp_samples_print_line}', file = hout)


def sort_bed(input_bed, output_bed, is_bedpe = False):

    with open(output_bed, 'w') as hout:
        if is_bedpe:
            subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", "-k4,4", "-k5,5n", "-k6,6n", input_bed], stdout = hout)
        else:
            subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", input_bed], stdout = hout)


def bgzip_tabix(input_bed, output_bedgz):

    with open(output_bedgz, 'w') as hout: 
        subprocess.check_call(["bgzip", "-f", "-c", input_bed], stdout = hout)
     
    subprocess.check_call(["tabix", "-p", "bed", output_bedgz])


def merge_control_from_parse_files(prefix_list_file, output_prefix):

    output_dir = os.path.dirname(output_prefix)
    if output_dir != '' and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    with open(prefix_list_file, 'r') as hin:
        for line in hin:
            prefix = line.rstrip('\n')
    
            if not os.path.exists(prefix + ".rearrangement.sorted.bedpe.gz"):
                logger.error(f'No file: {prefix}/.rearrangement.sorted.bedpe.gz')
                sys.exit(1)
            if not os.path.exists(prefix + ".insertion.sorted.bed.gz"):
                logger.error(f'No file: {prefix}/.insertion.sorted.bed.gz')
                sys.exit(1)
            if not os.path.exists(prefix + ".deletion.sorted.bed.gz"):
                logger.error(f'No file: {prefix}/.deletion.sorted.bed.gz')
                sys.exit(1)
            if not os.path.exists(prefix + ".bp_info.sorted.bed.gz"):
                logger.error(f'No file: {prefix}/.bp_info.sorted.bed.gz')
                sys.exit(1)

    hout_r = open(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp1", 'w')
    hout_i = open(output_prefix + ".insertion.sorted.bed.gz.tmp1", 'w')
    hout_d = open(output_prefix + ".deletion.sorted.bed.gz.tmp1", 'w')
    hout_b = open(output_prefix + ".bp_info.sorted.bed.gz.tmp1", 'w')
    prefix_ind = 0
    with open(prefix_list_file, 'r') as hin:
        for line in hin:
            prefix = line.rstrip('\n')
            prefix_label = os.path.basename(prefix) + '_' + str(prefix_ind)

            merge_bed_by_sv(prefix + ".rearrangement.sorted.bedpe.gz", hout_r, prefix_label, is_bedpe = True)
            merge_bed_by_sv(prefix + ".insertion.sorted.bed.gz", hout_i, prefix_label)
            merge_bed_by_sv(prefix + ".deletion.sorted.bed.gz", hout_d, prefix_label)
            merge_bed_by_sv(prefix + ".bp_info.sorted.bed.gz", hout_b, prefix_label)

            prefix_ind = prefix_ind + 1


    hout_r.close()
    hout_i.close()
    hout_d.close()
    hout_b.close()

    sort_bed(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp1", output_prefix + ".rearrangement.sorted.bedpe.gz.tmp2", is_bedpe = True)
    sort_bed(output_prefix + ".insertion.sorted.bed.gz.tmp1", output_prefix + ".insertion.sorted.bed.gz.tmp2")
    sort_bed(output_prefix + ".deletion.sorted.bed.gz.tmp1", output_prefix + ".deletion.sorted.bed.gz.tmp2")
    sort_bed(output_prefix + ".bp_info.sorted.bed.gz.tmp1", output_prefix + ".bp_info.sorted.bed.gz.tmp2")

    merge_bed_by_sample(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp2", output_prefix + ".rearrangement.sorted.bedpe.gz.tmp3", is_bedpe = True)
    merge_bed_by_sample(output_prefix + ".insertion.sorted.bed.gz.tmp2", output_prefix + ".insertion.sorted.bed.gz.tmp3")
    merge_bed_by_sample(output_prefix + ".deletion.sorted.bed.gz.tmp2", output_prefix + ".deletion.sorted.bed.gz.tmp3")
    merge_bed_by_sample(output_prefix + ".bp_info.sorted.bed.gz.tmp2", output_prefix + ".bp_info.sorted.bed.gz.tmp3")

    bgzip_tabix(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp3", output_prefix + ".rearrangement.sorted.bedpe.gz")
    bgzip_tabix(output_prefix + ".insertion.sorted.bed.gz.tmp3", output_prefix + ".insertion.sorted.bed.gz")
    bgzip_tabix(output_prefix + ".deletion.sorted.bed.gz.tmp3", output_prefix + ".deletion.sorted.bed.gz")
    bgzip_tabix(output_prefix + ".bp_info.sorted.bed.gz.tmp3", output_prefix + ".bp_info.sorted.bed.gz")

    os.remove(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp1")
    os.remove(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp2")
    os.remove(output_prefix + ".rearrangement.sorted.bedpe.gz.tmp3")
    os.remove(output_prefix + ".insertion.sorted.bed.gz.tmp1")
    os.remove(output_prefix + ".insertion.sorted.bed.gz.tmp2")
    os.remove(output_prefix + ".insertion.sorted.bed.gz.tmp3")
    os.remove(output_prefix + ".deletion.sorted.bed.gz.tmp1")
    os.remove(output_prefix + ".deletion.sorted.bed.gz.tmp2")
    os.remove(output_prefix + ".deletion.sorted.bed.gz.tmp3")
    os.remove(output_prefix + ".bp_info.sorted.bed.gz.tmp1")
    os.remove(output_prefix + ".bp_info.sorted.bed.gz.tmp2")
    os.remove(output_prefix + ".bp_info.sorted.bed.gz.tmp3")


